import argparse
import coloc_sat
from coloc_sat.generate_coloc import GenerateColoc
from coloc_sat.intersection import __version__
import sys


def main():
    parser = argparse.ArgumentParser(description="Generate co-locations between two products.")

    parser.add_argument("--product1-id", type=str, help="Path of the first product.")
    parser.add_argument("--product2-id", type=str, help="Path of the second product")
    parser.add_argument("--destination-folder", default='/tmp', nargs='?', type=str, help="Folder path for the output.")
    parser.add_argument("--delta-time", default=30, nargs='?', type=int,
                        help="Maximum time in minutes between two product acquisitions.")
    parser.add_argument("--minimal-area", default='1600km2', nargs='?', type=str,
                        help="Minimal intersection area in square kilometers.")
    parser.add_argument("--listing", default=True, action="store_true",
                        help="Create a listing of co-located files.")
    parser.add_argument("--no-listing", dest="listing", action="store_false", help="Do not create a listing.")
    parser.add_argument("--product-generation", default=True, action="store_true",
                        help="Generate a co-location product.")
    parser.add_argument("--no-product-generation", dest="product_generation", action="store_false",
                        help="Do not generate a co-location product.")
    parser.add_argument("--listing-filename", nargs='?', type=str,
                        help="Name of the listing file to be created.")
    parser.add_argument("--colocation-filename", nargs='?', type=str,
                        help="Name of the co-location product to be created.")
    parser.add_argument("--config", type=str, help="Configuration file to use instead of the "
                                                   "default one.")
    parser.add_argument("-v", "--version", action="store_true", help="Print version")

    args = parser.parse_args()

    if args.version:
        print(__version__)
        sys.exit(0)

    print(f"The script is executed from {__file__}")

    # Check for missing required arguments
    if not args.product1_id or not args.product2_id:
        parser.error("product1-id and product2-id are required arguments.")

    # Information for the user about listing / co-location product creation
    if args.listing is True:
        print("A listing of the co-located products will be created. To disable, use --no-listing.")
    if args.product_generation is True:
        print("Co-location products will be created. To disable, use --no-product-generation.")

    print("WARNING : product colocation has only been tested on _ll_gd SAR products.")

    generator = GenerateColoc(**vars(args))
    generator.save_results()

    print("Coloc python program successfully ended")
    


import os
import numpy as np
from datetime import datetime

from .tools import correct_dataset, convert_mingmt, common_var_names
from .windsat_daily_v7 import WindSatDaily, to_xarray_dataset


class GetWindSatMeta:
    def __init__(self, product_path, product_generation=False):
        self.product_path = product_path
        self.product_name = os.path.basename(self.product_path)
        self.product_generation = product_generation
        self._time_name = 'time'
        self._longitude_name = 'longitude'
        self._latitude_name = 'latitude'
        self._dataset = to_xarray_dataset(WindSatDaily(product_path, np.nan)).load()
        self.dataset = correct_dataset(self.dataset, self.longitude_name)
        self.dataset = convert_mingmt(self)

    @property
    def longitude_name(self):
        """
        Get the name of the longitude variable in the dataset

        Returns
        -------
        str
            longitude name
        """
        return self._longitude_name

    @property
    def latitude_name(self):
        """
        Get the name of the latitude variable in the dataset

        Returns
        -------
        str
            latitude name
        """
        return self._latitude_name

    @property
    def time_name(self):
        """
        Get the name of the time variable in the dataset

        Returns
        -------
        str
            time name
        """
        return self._time_name

    @property
    def dataset(self):
        """
        Getter for the acquisition dataset

        Returns
        -------
        xarray.Dataset
            Acquisition dataset
        """
        return self._dataset

    @dataset.setter
    def dataset(self, value):
        """
        Setter of attribute `self.dataset`

        Parameters
        ----------
        value: xarray.Dataset
            new Dataset
        """
        self._dataset = value

    @property
    def day_date(self):
        """
        Get day date from the product name as a datetime

        Returns
        -------
        datetime.datetime
        Day date of the product
        """
        str_date = self.product_name.split('_')[1].split('v')[0]
        return datetime.strptime(str_date, '%Y%m%d')

    @property
    def minute_name(self):
        """
        Get name of the minute variable in the dataset

        Returns
        -------
        str
            Minute variable name
        """
        return 'mingmt'

    @property
    def acquisition_type(self):
        """
        Gives the acquisition type (swath, truncated_swath,daily_regular_grid, model_regular_grid)

        Returns
        -------
        str
            acquisition type

        """
        return 'daily_regular_grid'

    @property
    def start_date(self):
        """
        Start acquisition time

        Returns
        -------
        numpy.datetime64
            Start time
        """
        return min(np.unique(self.dataset[self.time_name]))

    @property
    def stop_date(self):
        """
        Stop acquisition time

        Returns
        -------
        numpy.datetime64
            Stop time
        """
        return max(np.unique(self.dataset[self.time_name]))

    @property
    def orbit_segment_name(self):
        """
        Gives the name of the variable for orbit segmentation in dataset (Ascending / Descending). If value is None,
        so the orbit hasn't orbited segmentation

        Returns
        -------
        str | None
            Orbit segmentation variable name in the dataset. None if there isn't one.
        """
        return 'orbit_segment'

    @property
    def has_orbited_segmentation(self):
        """
        True if there is orbit segmentation in the dataset

        Returns
        -------
        bool
            Presence or not of an orbit segmentation
        """
        if self.orbit_segment_name is not None:
            return True
        else:
            return False

    @property
    def wind_name(self):
        """
        Name of an important wind variable in the dataset

        Returns
        -------
        str
            Wind variable name

        """
        return 'wdir'

    @property
    def mission_name(self):
        """
        Get the mission name (ex : RADARSAT-2, RCM, SENTINEL-1, SMOS, SMAP,...)

        Returns
        -------
        str
            Mission name
        """
        return 'WINDSAT'

    def rename_vars_in_coloc(self, dataset=None):
        """
        Rename variables from a dataset to homogenize the co-location product. If no dataset is explicit, so it is this
        of `self.dataset` which is used.

        Parameters
        ----------
        dataset: xarray.Dataset | None
            Dataset on which common vars must be renamed

        Returns
        -------
        xarray.Dataset
            Dataset with homogene variable names
        """
        if dataset is None:
            dataset = self.dataset
            # map the variable names in the dataset with the keys in common vars
        mapper = {
            self.wind_name: 'wind_direction',
            'w-mf': 'wind_speed',
        }
        for var in dataset.variables:
            if var in mapper.keys():
                key_in_common_vars = mapper[var]
                dataset = dataset.rename_vars({var: common_var_names[key_in_common_vars]})
        return dataset

    @property
    def unecessary_vars_in_coloc_product(self):
        """
        Get unecessary variables in co-location product

        Returns
        -------
        list[str]
            Unecessary variables in co-location product
        """
        return [self.time_name, 'land', 'nodata', 'ice', 'cloud']

    @property
    def necessary_attrs_in_coloc_product(self):
        """
        Get necessary dataset attributes in co-location product

        Returns
        -------
        list[str]
            Necessary dataset attributes in co-location product
        """
        # No attributes to the original dataset
        return []

    def rename_attrs_in_coloc_product(self, attr):
        """
        Get the new name of an attribute in co-location products from an original attribute

        Parameters
        ----------
        attr: str
            Attribute from the satellite dataset that needs to be renames for the co-location product.

        Returns
        -------
        str
            New attribute's name from the satellite dataset.
        """
        # No attributes to the original dataset
        return ""

    @longitude_name.setter
    def longitude_name(self, value):
        self._longitude_name = value

    @latitude_name.setter
    def latitude_name(self, value):
        self._latitude_name = value



import json
import math
from pathlib import Path

import numpy as np
import copy
import xarray as xr
import logging
import shapely
import rasterio

from .intersection_tools import extract_times_dataset, are_dimensions_empty, get_footprint_from_ll_ds, \
    get_polygon_area_in_km_squared, get_transform, get_common_points, get_nearest_time_datasets, remove_nat
from .tools import mean_time_diff, reformat_meta, convert_str_to_polygon
from .version import __version__


logger = logging.getLogger(__name__)


class ProductIntersection:
    def __init__(self, meta1, meta2, delta_time=60, minimal_area=1600, product_generation=True):
        # Store the meta and rename longitude/latitude by reference ones
        self._meta1 = reformat_meta(meta1)
        self._meta2 = reformat_meta(meta2)
        self.product_generation = product_generation
        self.delta_time = delta_time
        self.minimal_area = minimal_area
        self.delta_time_np = np.timedelta64(delta_time, 'm')
        self.start_date = None
        self.stop_date = None
        self._datasets = {}
        self.common_footprint = None
        self.resampled_datasets = None
        self.common_zone_datasets = None
        self.colocation_product = None

    @property
    def has_intersection(self):
        """
        Property that verifies if there is an intersection between 2 products.

        Returns
        -------
        bool
            True if 2 products are co-located.
        """
        times1 = (self.meta1.start_date - self.delta_time_np, self.meta1.stop_date + self.delta_time_np)
        times2 = (self.meta2.start_date - self.delta_time_np, self.meta2.stop_date + self.delta_time_np)
        if times1[1] < times2[0] or times2[1] < times1[0]:
            return False  # No time match => no footprint
        else:
            self.start_date = max(times1[0], times2[0])
            self.stop_date = min(times1[1], times2[1])
        if (self.meta1.acquisition_type == 'truncated_swath') \
                and (self.meta2.acquisition_type == 'truncated_swath'):
            fp1 = self.meta1.footprint
            fp2 = self.meta2.footprint
            is_intersected = fp1.intersects(fp2)
            if is_intersected:
                self.fill_common_footprint(fp1.intersection(fp2))
            return self._is_considered_as_intersected
        elif ((self.meta1.acquisition_type == 'truncated_swath') and
              (self.meta2.acquisition_type == 'daily_regular_grid')) or \
                ((self.meta2.acquisition_type == 'truncated_swath') and
                 (self.meta1.acquisition_type == 'daily_regular_grid')):
            return self.intersection_drg_truncated_swath()
        elif ((self.meta1.acquisition_type == 'truncated_swath') and
              (self.meta2.acquisition_type == 'swath')) or \
                ((self.meta2.acquisition_type == 'truncated_swath') and
                 (self.meta1.acquisition_type == 'swath')):
            return self.intersection_swath_truncated_swath()
        elif ((self.meta1.acquisition_type == 'model_regular_grid') or
              (self.meta2.acquisition_type == 'model_regular_grid')):
            return self.intersection_with_model()
        elif (self.meta1.acquisition_type == 'swath') and (self.meta2.acquisition_type == 'swath'):
            return self.intersection_non_truncated_swath_non_truncated_swath()
        elif ((self.meta1.acquisition_type == 'daily_regular_grid') and
              (self.meta2.acquisition_type == 'daily_regular_grid')):
            return self.intersection_drg_drg()
        elif ((self.meta1.acquisition_type == 'daily_regular_grid') and
              (self.meta2.acquisition_type == 'swath')) or \
                ((self.meta2.acquisition_type == 'daily_regular_grid') and
                 (self.meta1.acquisition_type == 'swath')):
            return self.intersection_drg_non_truncated_swath()

    def fill_common_footprint(self, footprint):
        if self.common_footprint is None:
            self.common_footprint = footprint
        else:
            self.common_footprint = self.common_footprint.union(footprint)

    @property
    def _is_considered_as_intersected(self):
        if self.common_footprint is None:
            return False
        else:
            area_intersection = get_polygon_area_in_km_squared(self.common_footprint)
            if area_intersection >= self.minimal_area:
                return True
            else:
                return False

    def intersection_with_model(self):
        """
        Method that verifies if there is an intersection between a product and a model (ERA 5 for example).
        This method can fill attributes like `self._datasets` (with the datasets after temporal and
        spatial intersection) and `self._common_footprint`.

        Returns
        -------
        bool
            True if there is an intersection (so if the products are co-located)
        """
        def rasterize_polygon(open_acquisition, polygon):
            if open_acquisition.acquisition_type == 'model_regular_grid':
                lon_name = open_acquisition.longitude_name
                lat_name = open_acquisition.latitude_name
                min_bounds = (min(np.unique(open_acquisition.dataset[lon_name])),
                              min(np.unique(open_acquisition.dataset[lat_name])))
                # we can get resolutions like this because it is a regular grid
                lon_res = abs(open_acquisition.dataset[lon_name][1] - open_acquisition.dataset[lon_name][0])
                lat_res = abs(open_acquisition.dataset[lat_name][1] - open_acquisition.dataset[lat_name][0])
                out_shape = [len(open_acquisition.dataset[lat_name]), len(open_acquisition.dataset[lon_name])]
                transform = rasterio.Affine.translation(min_bounds[0], min_bounds[1]) * rasterio.Affine.scale(lon_res,
                                                                                                              lat_res)
                return rasterio.features.rasterize(shapes=[polygon], out_shape=out_shape, transform=transform)
            else:
                raise ValueError('`rasterize_polygon` only can be applied on daily regular grid acquisition')

        def geographic_intersection(open_acquisition, polygon=None):
            if open_acquisition.acquisition_type == 'model_regular_grid':
                if polygon is None:
                    return open_acquisition.dataset
                else:
                    lon_name = open_acquisition.longitude_name
                    lat_name = open_acquisition.latitude_name

                    rasterized = rasterize_polygon(open_acquisition, polygon)
                    dataset = open_acquisition.dataset.where(rasterized)

                    dataset = dataset.dropna(lon_name, how='all')
                    dataset = dataset.dropna(lat_name, how='all')
                    return dataset
            else:
                raise ValueError('`geographic_intersection` only can be applied on model regular grid acquisition')

        if self.product_generation:
            if self.meta1.acquisition_type == 'model_regular_grid':
                model = self.meta1
                other_meta = self.meta2
            elif self.meta2.acquisition_type == 'model_regular_grid':
                model = self.meta2
                other_meta = self.meta1
            if other_meta.acquisition_type == 'model_regular_grid':
                unique_lon = np.unique(other_meta.dataset[other_meta.longitude_name])
                unique_lat = np.unique(other_meta.dataset[other_meta.latitude_name])
                corners = [(min(unique_lon), min(unique_lat)), (min(unique_lon), max(unique_lat)),
                           (max(unique_lon), max(unique_lat)), (max(unique_lon), min(unique_lat))]
                fp = shapely.geometry.Polygon(corners)
            else:
                fp = get_footprint_from_ll_ds(other_meta)
            # FIXME: for the moment we only arrive to get a dataset after spatial and temporal extraction when
            #  the co-location is with a truncated swath
            if other_meta.acquisition_type == 'truncated_swath':
                self._datasets[model.product_name] = geographic_intersection(model, polygon=other_meta.footprint)
            self.fill_common_footprint(fp)
        # if it is a model so there is data every day and worldwide => file can be co-located
        # (model file has been chosen depending on the date)
        return True

    def intersection_drg_truncated_swath(self):
        """
        Method that verifies if there is an intersection between a daily regular grid product and a truncated swath.
        This method can fill attributes like `self._datasets` (with the datasets after temporal and
        spatial intersection) and `self._common_footprint`.

        Returns
        -------
        bool
            True if there is an intersection (so if the products are co-located)
        """
        def rasterize_polygon(open_acquisition, polygon):
            if open_acquisition.acquisition_type == 'daily_regular_grid':
                lon_name = open_acquisition.longitude_name
                lat_name = open_acquisition.latitude_name
                min_bounds = (min(np.unique(open_acquisition.dataset[lon_name])),
                              min(np.unique(open_acquisition.dataset[lat_name])))
                # we can get resolutions like this because it is a regular grid
                lon_res = abs(open_acquisition.dataset[lon_name][1] - open_acquisition.dataset[lon_name][0])
                lat_res = abs(open_acquisition.dataset[lat_name][1] - open_acquisition.dataset[lat_name][0])
                out_shape = [len(open_acquisition.dataset[lat_name]), len(open_acquisition.dataset[lon_name])]
                transform = rasterio.Affine.translation(min_bounds[0], min_bounds[1]) * rasterio.Affine.scale(lon_res,
                                                                                                              lat_res)
                return rasterio.features.rasterize(shapes=[polygon], out_shape=out_shape, transform=transform)
            else:
                raise ValueError('`rasterize_polygon` only can be applied on daily regular grid acquisition')

        def geographic_intersection(open_acquisition, polygon=None):
            if open_acquisition.acquisition_type == 'daily_regular_grid':
                if polygon is None:
                    return open_acquisition.dataset
                else:
                    lon_name = open_acquisition.longitude_name
                    lat_name = open_acquisition.latitude_name

                    rasterized = rasterize_polygon(open_acquisition, polygon)
                    dataset = open_acquisition.dataset.where(rasterized)

                    dataset = dataset.dropna(lon_name, how='all')
                    dataset = dataset.dropna(lat_name, how='all')
                    return dataset
            else:
                raise ValueError('`geographic_intersection` only can be applied on daily regular grid acquisition')

        def spatial_temporal_intersection(open_acquisition, polygon=None):
            if open_acquisition.acquisition_type == 'daily_regular_grid':
                dataset = geographic_intersection(open_acquisition, polygon)
                dataset = extract_times_dataset(open_acquisition, dataset=dataset,
                                                start_date=self.start_date, stop_date=self.stop_date)
                return dataset.where(~np.isnan(dataset[open_acquisition.wind_name]), drop=True)
            else:
                raise ValueError(
                    '`spatial_temporal_intersection` only can be applied on daily regular grid acquisition')

        def verify_intersection(_ds):
            if (_ds is not None) and (not are_dimensions_empty(_ds)):
                poly = get_footprint_from_ll_ds(daily, _ds)
                is_intersected = poly.intersects(fp)
                if is_intersected:
                    self.fill_common_footprint(poly.intersection(fp))
                return self._is_considered_as_intersected
            else:
                return False

        if (self.meta1.acquisition_type == 'truncated_swath') and \
                (self.meta2.acquisition_type == 'daily_regular_grid'):
            truncated = self.meta1
            daily = self.meta2
        elif (self.meta2.acquisition_type == 'truncated_swath') and \
                (self.meta1.acquisition_type == 'daily_regular_grid'):
            truncated = self.meta2
            daily = self.meta1
        else:
            raise ValueError('intersection_drg_truncated_swath only can be used with a daily regular grid \
                                acquisition and a truncated one')
        fp = truncated.footprint
        if daily.has_orbited_segmentation:
            li = []
            # list that store booleans to express if an orbit has an intersection
            orbit_intersections = []
            for orbit in daily.dataset[daily.orbit_segment_name].data:
                sub_daily = copy.copy(daily)
                # Select orbit in the dataset of sub_daily
                sub_daily.dataset = sub_daily.dataset.sel(**{sub_daily.orbit_segment_name: orbit})
                _ds = spatial_temporal_intersection(sub_daily, polygon=fp)
                li.append(_ds.assign_coords(**{sub_daily.orbit_segment_name: orbit}))
                orbit_intersections.append(verify_intersection(_ds))

            self._datasets[daily.product_name] = xr.concat(li, dim=daily.orbit_segment_name)

            # if one of the orbit has an intersection, return True
            return any(orbit_intersections)
        else:
            _ds = spatial_temporal_intersection(daily, polygon=fp)
            self._datasets[daily.product_name] = _ds
            return verify_intersection(_ds)

    def intersection_swath_truncated_swath(self):
        """
        Method that verifies if there is an intersection between a truncated swath product and a non-truncated swath.
        This method can fill attributes like `self._datasets` (with the datasets after temporal and
        spatial intersection) and `self._common_footprint`.

        Returns
        -------
        bool
            True if there is an intersection (so if the products are co-located)
        """

        def geographic_intersection(open_acquisition, polygon=None):
            if open_acquisition.acquisition_type == 'swath':
                if polygon is None:
                    return open_acquisition.dataset
                else:
                    lon_name = open_acquisition.longitude_name
                    lat_name = open_acquisition.latitude_name

                    ds_scat = open_acquisition.dataset
                    # Find the scatterometer points that are within the sar swath bounding box
                    min_lon, min_lat, max_lon, max_lat = polygon.bounds
                    condition = (ds_scat[lon_name] > min_lon) & (ds_scat[lon_name] < max_lon) & \
                                (ds_scat[lat_name] > min_lat) & (ds_scat[lat_name] < max_lat)
                    ds_scat_intersected = ds_scat.where(condition, drop=True)
                    return ds_scat_intersected
            else:
                raise ValueError('`geographic_intersection` only can be applied on daily regular grid acquisition')

        def spatial_temporal_intersection(open_acquisition, polygon=None):
            if open_acquisition.acquisition_type == 'swath':
                dataset = geographic_intersection(open_acquisition, polygon)
                dataset = extract_times_dataset(open_acquisition, dataset=dataset,
                                                start_date=self.start_date, stop_date=self.stop_date)
                return dataset
            else:
                raise ValueError(
                    '`spatial_temporal_intersection` only can be applied on daily regular grid acquisition')

        def verify_intersection(swath_acquisition, footprint):
            # dataset where latitude and longitude are in the truncated swath footprint bounds,
            # and where time criteria is respected
            _ds = spatial_temporal_intersection(swath_acquisition, polygon=footprint)
            if (_ds is not None) and (not are_dimensions_empty(_ds)):
                """flatten_lon = _ds[swath_acquisition.longitude_name].data.flatten()
                flatten_lat = _ds[swath_acquisition.latitude_name].data.flatten()
                # Create a multipoint from swath lon/lat that are in the box and respect time criteria
                mpt = MultiPoint([(lon, lat) for lon, lat in zip(flatten_lon, flatten_lat)
                                  if (np.isfinite(lon) and np.isfinite(lat))])
                # Verify if a part of this multipoint can be intersected with the truncated swath footprint
                return mpt.intersects(footprint)"""
                poly = get_footprint_from_ll_ds(swath_acquisition, _ds)
                is_intersected = poly.intersects(footprint)
                if is_intersected:
                    self.fill_common_footprint(poly.intersection(footprint))
                return self._is_considered_as_intersected
            else:
                return False

        if (self.meta1.acquisition_type == 'truncated_swath') and \
                (self.meta2.acquisition_type == 'swath'):
            truncated = self.meta1
            swath = self.meta2
        elif (self.meta2.acquisition_type == 'truncated_swath') and \
                (self.meta1.acquisition_type == 'swath'):
            truncated = self.meta2
            swath = self.meta1
        else:
            raise ValueError('intersection_swath_truncated_swath only can be used with a swath \
                                acquisition and a truncated one')

        # footprint of the truncated swath
        fp = truncated.footprint

        if swath.has_orbited_segmentation:
            # list that store booleans to express if an orbit has an intersection
            orbit_intersections = []
            for orbit in swath.dataset[swath.orbit_segment_name]:
                sub_swath = copy.copy(swath)
                # Select orbit in the dataset of sub_daily
                sub_swath.dataset = sub_swath.dataset.sel(**{sub_swath.orbit_segment_name: orbit})
                orbit_intersections.append(verify_intersection(sub_swath, footprint=fp))
            # if one of the orbit has an intersection, return True
            return any(orbit_intersections)
        else:
            return verify_intersection(swath, footprint=fp)

    def intersection_drg_non_truncated_swath(self):
        """
        Method that verifies if there is an intersection between a daily regular grid product and a non-truncated swath.
        This method can fill attributes like `self._datasets` (with the datasets after temporal and
        spatial intersection) and `self._common_footprint`.

        Returns
        -------
        bool
            True if there is an intersection (so if the products are co-located)
        """
        raise NotImplementedError("This property isn't yet implemented")

    def intersection_drg_drg(self):
        """
        Method that verifies if there is an intersection between 2 daily regular grid products.
        This method can fill attributes like `self._datasets` (with the datasets after temporal and
        spatial intersection) and `self._common_footprint`.

        Returns
        -------
        bool
            True if there is an intersection (so if the products are co-located)
        """
        raise NotImplementedError("This property isn't yet implemented")

    def intersection_non_truncated_swath_non_truncated_swath(self):
        """
        Method that verifies if there is an intersection between 2 non-truncated swath products.
        This method can fill attributes like `self._datasets` (with the datasets after temporal and
        spatial intersection) and `self._common_footprint`.

        Returns
        -------
        bool
            True if there is an intersection (so if the products are co-located)
        """
        raise NotImplementedError("This property isn't yet implemented")

    @property
    def coloc_resample(self):
        """
        Resample 2 satellite datasets from `self.meta1`and `self.meta2`. If a dataset exists in `self._datasets`
        (it means that a meta dataset has been intersected temporally and spatially), so this one is chosen.
        Notes : it uses `rasterio.reproject_match` with a bi-linear resampling.

        Returns
        -------
        Dict[str, Union[xarray.Dataset, str]]
            Two first values of the dictionary are resampled datasets from meta1 and meta 2.  Last value is a string
            that precise which dataset of both has been reprojected.
        """
        logger.info("Starting resampling.")
        existing_dataset_keys = list(self._datasets.keys())
        logger.info("Getting datasets.")
        # Getting datasets
        if self.meta1.product_name in existing_dataset_keys:
            dataset1 = self._datasets[self.meta1.product_name]
        else:
            dataset1 = self.meta1.dataset
        logger.info("meta1 dataset opened.")
        if self.meta2.product_name in existing_dataset_keys:
            dataset2 = self._datasets[self.meta2.product_name]
        else:
            dataset2 = self.meta2.dataset
        logger.info("meta1 dataset opened.")

        # alias to metaobjects
        meta1 = self.meta1
        meta2 = self.meta2

        dataset1, dataset2 = get_nearest_time_datasets(meta1, dataset1, meta2, dataset2)

        # security to be sure that there is not orbit in the datasets dimensions
        dataset1 = remove_nat(meta1, dataset1)
        dataset2 = remove_nat(meta2, dataset2)

        # Set crs if not defined
        if dataset1.rio.crs is None:
            dataset1.rio.write_crs(4326, inplace=True)
        if dataset2.rio.crs is None:
            dataset2.rio.write_crs(4326, inplace=True)
        logger.info("Renaming datasets coordinates into xy.")
        # replace lon and lat with x and y (necessary to make reproject_match() working)
        dataset1 = dataset1.rename({meta1.longitude_name: 'x', meta1.latitude_name: 'y'})
        dataset2 = dataset2.rename({meta2.longitude_name: 'x', meta2.latitude_name: 'y'})

        logger.info("Done renaming datasets coordinates into xy.")

        pixel_spacing_lon1 = dataset1.coords["x"][1] - dataset1.coords["x"][0]
        pixel_spacing_lat1 = dataset1.coords["y"][1] - dataset1.coords["y"][0]
        pixel_spacing_lon2 = dataset2.coords["x"][1] - dataset2.coords["x"][0]
        pixel_spacing_lat2 = dataset2.coords["y"][1] - dataset2.coords["y"][0]

        logger.info(
            "Modifying dataset coordinates in range 0-360 if coordinates do not cross Greenwich Meridian (lon = 0).")
        if (dataset1.x[0].values < 0 and dataset1.x[len(dataset1.x) - 1].values > 180) or (
                dataset2.x[0].values < 0 and dataset2.x[len(dataset2.x) - 1].values > 180):
            meridian_datasets = True
            logger.info("datasets cross Greenwich Meridian, dataset coordinated will be modified after reprojection.")
        else:
            dataset1["x"] = dataset1["x"] % 360
            dataset2["x"] = dataset2["x"] % 360
            meridian_datasets = False
        logger.info("Done modifying dataset coordinates.")

        logger.info("Reprojecting dataset with higher resolution to make it match with the lower resolution one.")
        if pixel_spacing_lon1 * pixel_spacing_lat1 <= pixel_spacing_lon2 * pixel_spacing_lat2:
            # some product don't have the same resolution, so we apply the same resolution by specifying a resampling
            dataset1 = dataset1.rio.reproject_match(dataset2, resampling=rasterio.enums.Resampling.bilinear)
            reprojected_dataset = "dataset1"
            logger.info("dataset1 reprojected")
        else:
            # some product don't have the same resolution, so we apply the same resolution by specifying a resampling
            dataset2 = dataset2.rio.reproject_match(dataset1, resampling=rasterio.enums.Resampling.bilinear)
            reprojected_dataset = "dataset2"
            logger.info("dataset2 reprojected.")
        logger.info("Done reprojecting dataset.")

        if meridian_datasets:
            logger.info(
                "Modifying dataset coordinated in range 0-360 if coordinates cross Greenwich Meridian (lon = 0).")
            dataset1["x"] = dataset1["x"] % 360
            dataset2["x"] = dataset2["x"] % 360
            logger.info("Done modifying dataset coordinates.")

        logger.info("Renaming dataset coordinates into lon-lat.")
        dataset1 = dataset1.rename({'x': meta1.longitude_name, 'y': meta1.latitude_name})
        dataset2 = dataset2.rename({'x': meta2.longitude_name, 'y': meta2.latitude_name})
        logger.info("Done renaming dataset coordinates into lon-lat.")

        logger.info("Done resampling.")
        return {'meta1': dataset1, 'meta2': dataset2, 'reprojected_dataset': reprojected_dataset}

    @property
    def get_common_zone(self):
        """
        Search for common zone between two resampled datasets (located in `self.resampled_datasets`); and put these two
        dataset in this common zone.

        Returns
        -------
        xarray.Dataset, xarray.Dataset
            Resampled datasets located in a common zone (longitude and latitude).
        """
        logger.info("Starting getting common zone.")
        logger.info('Start getting resampled datasets')
        if self.resampled_datasets is None:
            self.fill_resampled_datasets()
        dataset1 = self.resampled_datasets['meta1']
        dataset2 = self.resampled_datasets['meta2']
        reprojected_dataset = self.resampled_datasets['reprojected_dataset']
        logger.info('Done getting resampled datasets')

        meta1 = self.meta1
        meta2 = self.meta2

        # transform polygons in range 0-360
        def shape360(lon, lat):
            """shapely shape to 0 360 (for shapely.ops.transform)"""
            orig_type = type(lon)
            lon = np.array(lon) % 360
            return tuple([orig_type(lon), lat])

        logger.info("Getting intersection of polygons.")
        poly_intersection = self.common_footprint
        logger.info("Done getting intersection of polygons.")

        logger.info("Modifying polygons coords range into 0-360.")
        poly_intersection = shapely.ops.transform(lambda x, y, z=None: shape360(x, y), poly_intersection)
        logger.info("Done modifying polygons coords range into 0-360.")

        logger.info("Calculating geometry_mask of reprojected dataset.")
        if reprojected_dataset == "dataset1":
            lon_name = meta1.longitude_name
            lat_name = meta1.latitude_name
            geometry_mask = rasterio.features.geometry_mask([poly_intersection],
                                                            out_shape=(dataset1[lat_name].shape[0],
                                                                       dataset1[lon_name].shape[0]),
                                                            transform=get_transform(dataset1, lon_name, lat_name),
                                                            invert=True, all_touched=True)
        elif reprojected_dataset == "dataset2":
            lon_name = meta2.longitude_name
            lat_name = meta2.latitude_name
            geometry_mask = rasterio.features.geometry_mask([poly_intersection],
                                                            out_shape=(dataset2[lat_name].shape[0],
                                                                       dataset2[lon_name].shape[0]),
                                                            transform=get_transform(dataset2, lon_name, lat_name),
                                                            invert=True, all_touched=True)
        logger.info("Done calculating geometry_mask of reprojected dataset.")
        logger.info("Applying geometry_mask on datasets.")
        # Keep only values in common zone
        dataset1_common_zone = dataset1.where(geometry_mask)
        dataset2_common_zone = dataset2.where(geometry_mask)
        logger.info("Done applying geometry_mask on datasets.")

        logger.info("Reshaping datasets to keep common zone.")
        # reshape to reduce dataset size (avoid having too much non-necessary nan)
        # find lon_min, lon_max, lat_min and lat_max
        lon2D, lat2D = np.meshgrid(dataset1_common_zone[meta1.longitude_name],
                                   dataset1_common_zone[meta1.latitude_name])
        lon2D[~geometry_mask] = np.nan
        lat2D[~geometry_mask] = np.nan
        # We need to round these values to avoid a bug which can occur when merging datasets
        lon_min = round(np.nanmin(lon2D), 6)
        lon_max = round(np.nanmax(lon2D), 6)
        lat_min = round(np.nanmin(lat2D), 6)
        lat_max = round(np.nanmax(lat2D), 6)
        # reshape
        dataset1_common_zone = dataset1_common_zone.where(
            (dataset1_common_zone[meta1.longitude_name] > lon_min) &
            (dataset1_common_zone[meta1.longitude_name] < lon_max) &
            (dataset1_common_zone[meta1.latitude_name] > lat_min) &
            (dataset1_common_zone[meta1.latitude_name] < lat_max), drop=True)
        dataset2_common_zone = dataset2_common_zone.where(
            (dataset2_common_zone[meta2.longitude_name] > lon_min) &
            (dataset2_common_zone[meta2.longitude_name] < lon_max) &
            (dataset2_common_zone[meta2.latitude_name] > lat_min) &
            (dataset2_common_zone[meta2.latitude_name] < lat_max), drop=True)
        dataset1_common_zone = dataset1_common_zone.assign_attrs({"polygon_common_zone": str(poly_intersection)})
        logger.info("Done reshaping datasets to keep common zone.")

        logger.info("Modifying datasets coords in range -180,180.")
        dataset1_common_zone = dataset1_common_zone.assign_coords(
            lon=(((dataset1_common_zone[meta1.longitude_name] + 180) % 360) - 180))
        dataset2_common_zone = dataset2_common_zone.assign_coords(
            lon=(((dataset2_common_zone[meta2.longitude_name] + 180) % 360) - 180))
        logger.info("Modifying datasets coords in range -180,180.")

        dataset1_common_zone = dataset1_common_zone.where(np.isfinite(dataset1_common_zone[meta1.time_name]), drop=True)
        dataset2_common_zone = dataset2_common_zone.where(np.isfinite(dataset2_common_zone[meta2.time_name]), drop=True)

        logger.info("Done getting common zone.")
        return dataset1_common_zone, dataset2_common_zone

    def fill_resampled_datasets(self):
        """
        Fills resampled datasets in the attribute `self.resampled_datasets`
        """
        self.resampled_datasets = self.coloc_resample

    def fill_common_zone_datasets(self):
        """
        Fills common zone datasets in the attribute `self.common_zone_datasets`
        """
        _tmp_dic = {}
        # put 2 datasets in their common zone
        dataset1_common_zone, dataset2_common_zone = self.get_common_zone
        _tmp_dic[self.meta1.product_name] = dataset1_common_zone.squeeze()
        _tmp_dic[self.meta2.product_name] = dataset2_common_zone.squeeze()
        self.common_zone_datasets = _tmp_dic

    def format_datasets(self):
        """
        Apply vars and attributes renaming to prepare common zone datasets for the co-location product and add
        missing attributes

        Returns
        -------
        Dict[str, xarray.Dataset]
            Formatted common zone datasets
        """

        def apply_attributes_changes(meta, ds):
            necessary_attrs = meta.necessary_attrs_in_coloc_product
            attrs_rename_func = meta.rename_attrs_in_coloc_product
            # Only keep required attributes and rename these for the co-location product
            ds.attrs = {attrs_rename_func(attr): ds.attrs[attr] for attr in necessary_attrs if attr in ds.attrs}
            ds.attrs['sourceProduct'] = meta.product_name
            ds.attrs['missionName'] = meta.mission_name
            if meta.acquisition_type == 'truncated_swath':
                # if the acquisition is a truncated swath, the dataset (so the footprint) is already time selective
                footprint = meta.footprint
            else:
                footprint = get_footprint_from_ll_ds(meta, ds, self.start_date, self.stop_date)
            unique_time = np.unique(ds[meta.time_name])
            ds.attrs['measurementStartDate'] = str(min(unique_time))
            ds.attrs['measurementStopDate'] = str(max(unique_time))
            ds.attrs['footprint'] = str(footprint)
            return ds

        def only_keep_required_vars(meta, ds):
            unecessary_vars = meta.unecessary_vars_in_coloc_product
            ds = ds.drop_vars([var for var in ds.variables if var in unecessary_vars])
            return ds

        def rename_common_vars(meta, ds):
            if hasattr(meta, 'rename_vars_in_coloc'):
                ds = meta.rename_vars_in_coloc(ds)
            return ds

        def rename_vars_and_attributes_with_nb(ds, ds_nb):
            for var in ds.data_vars:
                ds = ds.rename_vars({var: f"{var}_{ds_nb}"})
            attributes = ds.attrs
            ds.attrs = {f"{attr}_{ds_nb}": ds.attrs[attr] for attr in attributes}
            return ds

        if self.common_zone_datasets is None:
            self.fill_common_zone_datasets()
        _tmp_dic = {}

        product_name1 = self.meta1.product_name
        dataset1 = self.common_zone_datasets[product_name1]
        product_name2 = self.meta2.product_name
        dataset2 = self.common_zone_datasets[product_name2]
        # can't format datasets and create co-location product if dimensions are empty
        if are_dimensions_empty(dataset1) or are_dimensions_empty(dataset2):
            raise ValueError('There are not enough common points to create the co-location product or format datasets')
        # Attributes changes
        dataset1 = apply_attributes_changes(self.meta1, dataset1)
        dataset2 = apply_attributes_changes(self.meta2, dataset2)
        # Variable selection
        dataset1 = only_keep_required_vars(self.meta1, dataset1)
        dataset2 = only_keep_required_vars(self.meta2, dataset2)
        # rename common vars
        dataset1 = rename_common_vars(self.meta1, dataset1)
        dataset2 = rename_common_vars(self.meta2, dataset2)
        # Keep only common points between the 2 datasets for common variables
        dataset1, dataset2 = get_common_points(dataset1, dataset2)
        # Add the dataset number in the variable and attributes name
        dataset1 = rename_vars_and_attributes_with_nb(dataset1, 1)
        dataset2 = rename_vars_and_attributes_with_nb(dataset2, 2)
        return dataset1, dataset2

    @property
    def merge_datasets(self):
        """
        Merge 2 formatted common zones datasets in an only dataset

        Returns
        -------
        xr.Dataset
            2 formatted common zones datasets merged
        """
        dataset1, dataset2 = self.format_datasets()

        def poly_common_zone():
            fp1 = convert_str_to_polygon(dataset1.attrs['footprint_1'])
            fp2 = convert_str_to_polygon(dataset2.attrs['footprint_2'])
            return str(fp1.intersection(fp2))

        def ws_analysis_attributes():
            def calculate_scatter_index(dataarray1, dataarray2):
                # calculate the scatter index with the wind values
                # calcultate RMSE first
                MSE = np.square(np.subtract(dataarray1, dataarray2)).mean()
                RMSE = math.sqrt(MSE)

                # calculate mean value of all observations
                obs = xr.concat([dataarray1, dataarray2], dim="z")
                mean_obs = np.mean(obs).values

                # calculate scatter index in percentage
                scatter_index = RMSE / mean_obs * 100
                return scatter_index

            # number of points used for calculation of bias and standard deviation
            sum_wind_speed = dataset1["wind_speed_1"] + dataset2["wind_speed_2"]
            # when additioning or substracting a value with a nan, it becomes a nan
            counted_points = np.count_nonzero(~np.isnan(sum_wind_speed))

            # Determine informations analysis for the wind speed
            dict_ws_analysis = {
                'counted_points': counted_points,
                'vmax_m_s': sum_wind_speed.max().item()
            }
            if counted_points > 100:
                # Determine the bias  # (m/s)
                # https://numpy.org/doc/stable/reference/generated/numpy.nanmean.html#numpy.nanmean
                dict_ws_analysis["Bias"] = np.nanmean(dataset1["wind_speed_1"] - dataset2["wind_speed_2"])
                # Determine the standard deviation  # (m/s)
                # https://numpy.org/doc/stable/reference/generated/numpy.nanstd.html#numpy.nanstd
                dict_ws_analysis["Standard deviation"] = np.nanstd(
                    dataset1["wind_speed_1"] - dataset2["wind_speed_2"])

                # Compute the Pearson correlation coefficient between two DataArray objects along a shared dimension.
                correlation_coefficient = xr.corr(dataset1["wind_speed_1"], dataset2["wind_speed_2"]).values
            else:
                dict_ws_analysis["Bias"] = 0
                dict_ws_analysis["Standard deviation"] = 0
                correlation_coefficient = 0

            dict_ws_analysis['scatter_index'] = calculate_scatter_index(dataset1["wind_speed_1"],
                                                                        dataset2["wind_speed_2"])
            return dict_ws_analysis

        def get_common_attrs():
            attrs = {}
            start1, stop1 = dataset1.attrs['measurementStartDate_1'], dataset1.attrs['measurementStopDate_1']
            start2, stop2 = dataset2.attrs['measurementStartDate_2'], dataset2.attrs['measurementStopDate_2']
            attrs['time_difference'] = str(mean_time_diff(start1, stop1, start2, stop2))
            attrs['polygon_common_zone'] = poly_common_zone()
            attrs['area_intersection'] = str(get_polygon_area_in_km_squared(attrs['polygon_common_zone']))
            # add tool version to attributes
            attrs['coloc_sat_version'] = __version__
            return attrs

        merged_ds = xr.merge([dataset1, dataset2], compat='override')
        merged_ds.attrs |= get_common_attrs()
        merged_ds.attrs |= dataset1.attrs
        merged_ds.attrs |= dataset2.attrs
        merged_ds.attrs |= ws_analysis_attributes()
        return merged_ds

    @property
    def coloc_product_datasets(self):
        """
        Get the final co-located product dataset.

        Notes:
            - This method also populates the common zone datasets in the `self.common_zone_datasets` attribute when they do not exist yet.
            - It also populates the resampled datasets in the `self.resampled_datasets` attribute when they do not exist yet.
            - It also formats the datasets (variable and attribute names + adds new attributes).
            - Finally, it stores the result in `self.colocation_product`.

        Returns:
        --------
        Dict[str, xarray.Dataset]
            Final co-located product datasets.
        """
        if self.colocation_product is None:
            self.colocation_product = self.merge_datasets
        return self.colocation_product

    @property
    def meta1(self):
        """
        Getter of the first metaobject.
        """
        return self._meta1

    @property
    def meta2(self):
        """
        Getter of the second metaobject.
        """
        return self._meta2

    @meta1.setter
    def meta1(self, value):
        """
        Setter of the first metaobject.
        """
        self._meta1 = value

    @meta2.setter
    def meta2(self, value):
        """
        Getter of the second metaobject.
        """
        self._meta2 = value



import os.path

from .tools import call_meta_class, get_all_comparison_files, extract_name_from_meta_class, set_config
from .intersection import ProductIntersection
from .sar_meta import GetSarMeta
import numpy as np


class GenerateColoc:
    """
        Class that generates co-locations. It can create listings of co-located products and/or generate co-location products.
        Some arguments of this class are expressed as keyword arguments because there are 2 use options.
        For the first option (comparison between a product and a whole dataset), arguments are `ds_name`, `input_ds`,
        `level`.
        For the second option (comparison between 2 products), argument is `product2_id`.

        Parameters
        ----------
        product1_id : str
            Path of a product for which we want to create a listing of its co-located products and/or generate a co-location product.
            If it is a SAR Level-1 product, only a listing of its co-located files will be done.
        destination_folder: str
            Folder path where listing and / or co-location products will be created
        delta_time : int
            Maximum time (in minutes) that can separate two product acquisitions.
        minimal_area : int | str
            Minimal intersection area restricted for a valid co-location. If it is an integer, so it is expressed in
            square kilometers. If it is a string, so the unit can be expressed as follows: `1600km2` or `1600000000m2`.
        listing: bool
            True if a listing of the co-located_files must be created. Default value is False
        product_generation: bool
            True if a co-location product must be created. Default value is True

        Keyword Arguments
        -----------------
        - For the first option (comparison between a product and a whole dataset):
            ds_name : str | None
                Name of the dataset to be compared. Choices can be 'S1', 'RS2', 'RCM', 'HY2', 'ERA5', 'WS', 'SMOS', 'SMAP'.
            input_ds : str | list[str] | None, optional
                Optional. Used if it is needed to compare with a subset of products. This subset can be a subset of product paths
                or a text file that contains the different paths. If not specified, the default value is None.
                NOTE: The subset of products must belong to the mission specified in the `ds_name` argument.
            level : int | None, optional
                When `ds_name` is SAR, specify the value of the product level. If it is None, get all SAR levels.
                It is useless to give it a value when `ds_name` is something other than a SAR ('S1', 'RS2', 'RCM').
                Values can be 1, 2, or None (default value).

        - For the second option (comparison between 2 products):
            product2_id : str | None
                Path of the product that must be compared with `product1`.

        Optional Arguments
        ------------------
        listing_filename : str | None, optional
            Name of the listing file that must be created. It is useless to specify one if `listing` is False.
            Default value is None.
        colocation_filename : str | None, optional
            Name of the co-location product that must be created. It is useless to specify one if `product_generation` is False.
            Default value is None.
        config : str | None, optional
            Path to configuration file to use. If not provided, the one located in ~/coloc_sat/localconfig.yaml will
            be used if it exists, else the config.yml of this package is used.
    """

    def __init__(self, product1_id, destination_folder='/tmp', delta_time=60, minimal_area=1600, listing=False,
                 product_generation=True, **kwargs):
        config_path = kwargs.get('config', None)
        if config_path is not None:
            set_config(config_path)
        # Define descriptive attributes
        self.level = kwargs.get('level', None)
        self.ds_name = kwargs.get('ds_name', None)
        self.input_ds = kwargs.get('input_ds', None)
        self._product_generation = product_generation
        self._listing = listing
        self.product1_id = product1_id
        self.product1 = call_meta_class(self.product1_id, product_generation=self._product_generation)
        self.product2_id = kwargs.get('product2_id', None)
        if self.product2_id is not None:
            self.product2 = call_meta_class(self.product2_id, product_generation=self._product_generation)
        else:
            self.product2 = None
        self.delta_time = delta_time
        self._minimal_area = minimal_area
        self.delta_time_np = np.timedelta64(delta_time, 'm')
        self.destination_folder = destination_folder
        self._listing_filename = kwargs.get('listing_filename', None)
        self._colocation_filename = kwargs.get('colocation_filename', None)
        # define other attributes
        self.comparison_files = self.get_comparison_files
        self.intersections = None
        self.colocated_files = None
        self.fill_intersections()
        self.fill_colocated_files()

    @property
    def minimal_area(self):
        """
        Get minimal intersection area restricted for a valid co-location. Expressed in square kilometers.

        Returns
        -------
        int
            Minimal area intersection in square kilometers
        """
        if isinstance(self._minimal_area, int):
            return self._minimal_area
        elif isinstance(self._minimal_area, str):
            if self._minimal_area.endswith('km2'):
                return int(self._minimal_area.replace('km2', ''))
            elif self._minimal_area.endswith('m2'):
                return int(self._minimal_area.replace('m2', '')) * 1e6
            else:
                raise ValueError('minimal_area expressed as a string in argument must end by km2 or m2')
        else:
            raise TypeError("minimal_area expressed as an argument must be a string or an integer. Please refer to " +
                            "the documentation")

    @property
    def compare2products(self):
        """
        Know if the comparison is between 2 products or not

        Returns
        -------
        bool
            True if the comparison is between 2 products
        """
        if (self.product2 is not None) and (self.ds_name is None):
            return True
        elif (self.product2 is None) and (self.ds_name is None):
            raise self.UnknownOptionError("The option hasn't been recognized. Please look at the doc string in the " +
                                          "code source of the `GenerateColoc` class. A value must be given to the +"
                                          "argument `product2` or to `ds_name`.")
        elif (self.product2 is not None) and (self.ds_name is not None):
            raise self.UnknownOptionError("The option hasn't been recognized. Please look at the doc string in the " +
                                          "code source of the `GenerateColoc` class. A value must be given to the +"
                                          "argument `product2` or to `ds_name`; NOT BOTH.")
        else:
            return False

    @property
    def listing(self):
        """
        Know if a listing must be created

        Returns
        -------
        bool
            True if a listing must be created
        """
        return self._listing

    def product_generation(self, intersection):
        """
        Know if a co-location product must be created

        Returns
        -------
        bool
            True if a co-location product must be created
        """
        meta1 = intersection.meta1
        meta2 = intersection.meta2
        generation_bool = self._product_generation
        if isinstance(meta1, GetSarMeta):
            if meta1.is_safe:
                generation_bool = False
        if isinstance(meta2, GetSarMeta):
            if meta2.is_safe:
                generation_bool = False
        return generation_bool

    def listing_filename(self, intersection):
        """
        Get the filename of the listing file that must be created

        Parameters
        ----------
        intersection: sar_coloc.ProductIntersection
            intersection between 2 products

        Returns
        -------
        str
            Filename of the listing file that must be created
        """
        if self._listing_filename is not None:
            return self._listing_filename
        else:
            meta1_name = extract_name_from_meta_class(intersection.meta1).upper()
            meta2_name = extract_name_from_meta_class(intersection.meta2).upper()
            return f"listing_coloc_{meta1_name}_{meta2_name}_{self.delta_time}.txt"

    def colocation_filename(self, intersection):
        """
        Get the filename of the co-location product that must be created

        Parameters
        ----------
        intersection: sar_coloc.ProductIntersection
            intersection between 2 products

        Returns
        -------
        str
            Filename of the co-location product that must be created
        """
        if self._colocation_filename is not None:
            return self._colocation_filename
        else:
            name1 = intersection.meta1.product_name.split('.')[0]
            name2 = intersection.meta2.product_name.split('.')[0]
            return f"sat_coloc_{name1}__{name2}.nc"

    @property
    def product1_start_date(self):
        """
        Get start date of the product1 considering the delta time

        Returns
        -------
        numpy.datetime64
            Start date of the product1 considering the delta time
        """
        return self.product1.start_date - self.delta_time_np

    @property
    def product1_stop_date(self):
        """
        Get stop date of the product1 considering the delta time

        Returns
        -------
        numpy.datetime64
            stop date of the product1 considering the delta time
        """
        return self.product1.stop_date + self.delta_time_np

    @property
    def get_comparison_files(self):
        """
        Get all the files from the specified database that match with the start and stop dates

        Returns
        -------
        list | None
            Comparison files.
        """
        if self.compare2products:
            return [self.product2_id]
        else:
            all_comparison_files = get_all_comparison_files(self.product1_start_date, self.product1_stop_date,
                                                            ds_name=self.ds_name, input_ds=self.input_ds,
                                                            level=self.level)
            if self.product1_id in all_comparison_files:
                all_comparison_files.remove(self.product1_id)
            return all_comparison_files

    def fill_intersections(self):
        """
        Fill a dictionary as `self.intersections` with intersections (`sar_coloc.ProductIntersection`) between
        `self.product1_id` and products that are in `self.comparison_files`. If no products are in
        `self.comparison_files`, so `self.intersections` remains with None value.
        """
        _intersections = {}
        for file in self.comparison_files:
            try:
                opened_file = call_meta_class(file, product_generation=self._product_generation)
                intersecter = ProductIntersection(self.product1, opened_file, delta_time=self.delta_time,
                                                  minimal_area=self.minimal_area,
                                                  product_generation=self._product_generation)
                _intersections[file] = intersecter
            except FileNotFoundError:
                pass
        if len(list(_intersections.keys())) > 0:
            self.intersections = _intersections

    def fill_colocated_files(self):
        """
        Fill a dictionary as `self.colocated_files` with file paths of products from `self.comparison_files` that can be
        colocated with `self.product1_id`. If no products are in `self.comparison_files`, so `self.colocated_files`
        remains with None value.
        """
        if self.intersections is not None:
            _colocated_files = []
            for filename, intersection in self.intersections.items():
                if intersection.has_intersection:
                    _colocated_files.append(filename)
            if len(_colocated_files) > 0:
                self.colocated_files = _colocated_files

    @property
    def has_coloc(self):
        """
        Know if the product `self.product1_id` has co-located products in `self.comparison_files`

        Returns
        -------
        bool
            True if the product has co-located products
        """
        if self.colocated_files is None:
            return False
        else:
            return True

    class UnknownOptionError(Exception):
        """
        Used to raise errors concerning the 2 arguments subsets given in input of the class `GenerateColoc`
        """
        pass

    def save_results(self):
        """
        Save the result listing as a text file, and / or the resulting co-location product as a netcdf file.
        The creation depends on the value of the attributes `self.colocated_files` and
        `self.product_generation(intersection)` for a specific intersection. Paths where files are written is specified
        in `self.listing_filename(intersection)` and `self.product_generation(intersection)`.
        """
        if self.colocated_files:
            for colocated_file in self.colocated_files:
                intersection = self.intersections[colocated_file]
                if self.listing:
                    listing_path = os.path.join(self.destination_folder, self.listing_filename(intersection))
                    # Create the destination directory if it doesn't exist
                    if not os.path.exists(self.destination_folder):
                        os.makedirs(self.destination_folder)
                    line = f"{self.product1.product_path}:{colocated_file}\n"
                    reversed_line = f"{colocated_file}:{self.product1.product_path}\n"
                    # only write the 2 co-located product if the co-location doesn't exist in the listing file
                    if os.path.exists(listing_path):
                        with open(listing_path, 'r') as listing_file:
                            existing_lines = listing_file.readlines()
                    else:
                        # if the listing file doesn't exist, so there are no existing lines
                        existing_lines = []
                    if (line not in existing_lines) and (reversed_line not in existing_lines):
                        with open(listing_path, 'a') as listing_file:
                            listing_file.write(line)
                        print(f"A co-located product have been added in the listing file " +
                              f"{listing_path}")
                    else:
                        print("A co-located product already exists in the listing file " +
                              f"{listing_path}")

                if self.product_generation(intersection):
                    colocation_product_path = os.path.join(self.destination_folder, self.colocation_filename(intersection))
                    # Create the destination directory if it doesn't exist
                    if not os.path.exists(self.destination_folder):
                        os.makedirs(self.destination_folder)
                    coloc_ds = intersection.coloc_product_datasets
                    coloc_ds.to_netcdf(colocation_product_path)
                    print(f"A co-located products have been created: {colocation_product_path}")
        else:
            print("No coloc file has been produced, probably because no coloc has been found.")



import numpy as np
import pyproj
from shapely import MultiPolygon
from shapely.geometry import Polygon, MultiPoint, LineString, Point
from itertools import product
import math
from affine import Affine
from .tools import extract_name_from_meta_class, convert_str_to_polygon


def extract_times_dataset(acquisition, dataset=None, start_date=None, stop_date=None):
    """
    Extract a sub-dataset from a dataset of an acquisition to get a time dataset within 2 bounds (dates). If one of
    th bound exceeds the acquisition extremum times, so the acquisition Start and/ or Stop dates are chosen.

    Parameters
    ----------
    acquisition: coloc_sat.GetSarMeta | coloc_sat.GetSmosMeta | coloc_sat.GetEra5Meta | coloc_sat.GetHy2Meta | coloc_sat.GetSmapMeta | coloc_sat.GetWindsatMeta
        Meta object from an acquisition
    dataset: xarray.Dataset | None
        dataset on which we want to extract some values depending on times. If it is None, extraction is made on
        `open_acquisition.dataset`; else extraction is made on the specified dataset
    start_date: numpy.datetime64 | None
        Start chosen date.
    stop_date: numpy.datetime64 | None
        End chosen date.

    Returns
    -------
    xarray.Dataset | None
        Contains a sub-dataset of the acquisition dataset (between `start_date` and `stop_date`).
    """
    if dataset is None:
        dataset = acquisition.dataset
    if dataset is None:
        return dataset
    if start_date is None:
        start_date = acquisition.start_date
    if stop_date is None:
        stop_date = acquisition.stop_date
    time_name = acquisition.time_name
    return dataset.where((dataset[time_name] >= start_date) &
                         (dataset[time_name] <= stop_date), drop=True)


def are_dimensions_empty(dataset):
    """
    Verify if a dataset has all its dimensions empty

    Parameters
    ----------
    dataset: xarray.Dataset
        dataset on which empty verification needs to be done

    Returns
    -------
    bool
        True if dataset has all its dimensions empty
    """
    for dimension in dataset.dims:
        if len(dataset[dimension]) != 0:
            return False  # One dimension isn't empty => there are values
    return True


def get_polygon_area_in_km_squared(polygon):
    """
    From a polygon, get its area in square kilometers

    Parameters
    ----------
    polygon: shapely.geometry.polygon.Polygon | shapely.geometry.multipolygon.MultiPolygon | str
        Shapely polygon (footprint for example)

    Returns
    -------
    float
        Area of the polygon in square kilometers
    """
    if isinstance(polygon, str):
        polygon = convert_str_to_polygon(polygon)

    # Define the projection for converting latitude/longitude to meters (EPSG:4326 -> EPSG:3857)
    proj = pyproj.Transformer.from_crs("EPSG:4326", "EPSG:3857", always_xy=True)

    if isinstance(polygon, Polygon):
        projected_polygon = Polygon(proj.itransform(polygon.exterior.coords))
        area_in_square_meters = projected_polygon.area

    elif isinstance(polygon, MultiPolygon):
        area_in_square_meters = sum(Polygon(proj.itransform(p.exterior.coords)).area for p in polygon.geoms)
    elif isinstance(polygon, LineString) or isinstance(polygon, Point):
        return 0.0

    else:
        raise ValueError(f"Area from type {type(polygon)} can't be computed. Polygon is : {polygon}")

    # Convert the area from square meters to square kilometers
    area_in_square_km = area_in_square_meters / 1e6

    return area_in_square_km


def get_footprint_from_ll_ds(acquisition, ds=None, start_date=None, stop_date=None):
    """
    Get the footprint from a dataset in an acquisition. If there is a start and a stop time, the footprint is selected
    on this time range.

    Parameters
    ----------
    acquisition: coloc_sat.GetSarMeta | coloc_sat.GetSmosMeta | coloc_sat.GetEra5Meta | coloc_sat.GetHy2Meta | coloc_sat.GetSmapMeta | coloc_sat.GetWindsatMeta
        Meta acquisition.
    ds: xarray.Dataset
        Ds on which footprint can be extracted from longitude and latitude vars. If no one is given (value at None)
        , dataset is taken from `acquisition.dataset`.
    start_date: numpy.datetime64 | None
        Start chosen date.
    stop_date: numpy.datetime64 | None
        End chosen date.

    Returns
    -------
    shapely.geometry.polygon.Polygon
        Footprint of the dataset
    """
    if ds is None:
        ds = acquisition.dataset
    if (start_date is not None) or (stop_date is not None):
        ds = extract_times_dataset(acquisition, dataset=ds, start_date=start_date, stop_date=stop_date)
    flatten_lon = ds[acquisition.longitude_name].data.flatten()
    flatten_lat = ds[acquisition.latitude_name].data.flatten()
    mpt_coords = [(lon, lat) for lon, lat in product(flatten_lon, flatten_lat) if not (math.isnan(lon) or
                                                                                       math.isnan(lat))]
    return MultiPoint(mpt_coords).convex_hull


def get_transform(ds, lon_name, lat_name):
    pixel_spacing_lon = ds.coords[lon_name][1] - ds.coords[lon_name][0]
    pixel_spacing_lat = ds.coords[lat_name][1] - ds.coords[lat_name][0]

    transform = Affine(pixel_spacing_lon, 0.0, ds.coords[lon_name][0].values,
                       0.0, pixel_spacing_lat, ds.coords[lat_name][0].values)
    return transform


def get_common_points(dataset1, dataset2):
    # Obtenir les noms des variables présentes dans les deux datasets
    common_variable_names = set(dataset1.data_vars.keys()) & set(dataset2.data_vars.keys())

    # Créer un masque pour chaque variable ayant le même nom dans les deux datasets
    for variable_name in common_variable_names:
        mask1 = ~dataset1[variable_name].isnull()
        mask2 = ~dataset2[variable_name].isnull()
        common_mask = mask1 & mask2

        # Appliquer le masque à chaque dataset
        dataset1[variable_name] = dataset1[variable_name].where(common_mask)
        dataset2[variable_name] = dataset2[variable_name].where(common_mask)

    return dataset1, dataset2


def get_nearest_time_datasets(meta1, dataset1, meta2, dataset2):
    if (extract_name_from_meta_class(meta1) == 'Era5') or (extract_name_from_meta_class(meta2) == 'Era5'):
        if len(dataset1.time) > 1 and len(dataset2.time == 1):
            nearest_time = min(dataset1.time.data, key=lambda x: abs(x - dataset2.time.data[0]))
            dataset1 = dataset1.sel(time=nearest_time).squeeze()
        elif len(dataset2.time) > 1 and len(dataset1.time == 1):
            nearest_time = min(dataset2.time.data, key=lambda x: abs(x - dataset1.time.data[0]))
            dataset2 = dataset2.sel(time=nearest_time).squeeze()
    return dataset1, dataset2


def remove_nat(meta, dataset=None):
    """
    Remove not a time values in the variable time from the specified dataset. If there is an orbit_segment that is not
    used by variables after removing the NaT values, so it is removed.

    Parameters
    ----------
    meta: coloc_sat.GetSarMeta | coloc_sat.GetSmosMeta | coloc_sat.GetEra5Meta | coloc_sat.GetHy2Meta | coloc_sat.GetSmapMeta | coloc_sat.GetWindsatMeta
        Meta acquisition
    dataset: xarray.Dataset | None
        dataset from the acquisition on which the operation must be applied

    Returns
    -------
    xarray.Dataset
        Dataset without uselessness not a time values in the time variable

    Notes
    -----
    This function can be useful for a coloc after the spatial and temporal extraction to be sure that it doesn't
    remain values of orbit_segment.
    """
    if dataset is None:
        dataset = meta.dataset
    dataset = dataset.where(np.isfinite(dataset[meta.time_name]), drop=True).squeeze()
    if meta.has_orbited_segmentation:
        dimension_to_check = meta.orbit_segment_name
        # Verify if the orbit dimension is used by variables in the dataset
        used_by_variables = []
        for var_name in dataset.variables:
            var = dataset[var_name]
            if dimension_to_check in var.dims:
                used_by_variables.append(var_name)
        if not used_by_variables:
            dataset = dataset.drop_vars(meta.orbit_segment_name)
    return dataset



import os
from .tools import open_nc, correct_dataset, parse_date, common_var_names


class GetEra5Meta:
    def __init__(self, product_path, product_generation=False):
        self.product_path = product_path
        self.product_name = os.path.basename(self.product_path)
        self.product_generation = product_generation
        self._dataset = None
        # These attributes will be defined when the dataset will be reformatted for the coloc product generation
        # (see self.reformat_meta)
        self._longitude_name = None
        self._latitude_name = None

        if self.product_generation:
            self._dataset = open_nc(product_path).load()
            self.dataset = correct_dataset(self.dataset, lon_name=self.longitude_name_res(0.25))
            self.dataset = correct_dataset(self.dataset, lon_name=self.longitude_name_res(0.5))
            self.reformat_meta()

    @property
    def start_date(self):
        """
        Start acquisition time

        Returns
        -------
        numpy.datetime64
            Start time
        """
        # first time is at 00:00:00
        str_time = self.product_path.split('_')[-1].split('.')[0] + '000000'
        return parse_date(str_time)

    @property
    def stop_date(self):
        """
        Stop acquisition time

        Returns
        -------
        numpy.datetime64
            Stop time
        """
        # last time is at 23:00:00
        str_time = self.product_path.split('_')[-1].split('.')[0] + '230000'
        return parse_date(str_time)

    def longitude_name_res(self, resolution):
        """
        Get the name of the longitude variable in the dataset. For ERA 5, two longitude variable exist :
        one with a resolution of 0.25; and one with a resolution of 0.5

        Parameters
        ----------
        resolution: float
            Specified resolution for the dimension (dimension must exist in the dataset with the name
            `'longitude%s' % (str(resolution).replace('.', ''))` )

        Returns
        -------
        str
            longitude name
        """
        str_resolution = str(resolution).replace('.', '')
        str_resolution += '0' * (3 - len(str_resolution))  # Add 0 to have a str of 3 characters
        name = f"longitude{str_resolution}"
        if name in self.dataset.dims:
            return name
        else:
            raise ValueError(f"{name} wasn't found in the dataset. Please verify the resolution is correct")

    def latitude_name_res(self, resolution):
        """
        Get the name of the latitude variable in the dataset. For ERA 5, two latitude variable exist :
        one with a resolution of 0.25; and one with a resolution of 0.5

        Parameters
        ----------
        resolution: float
            Specified resolution for the dimension (dimension must exist in the dataset with the name
            `'latitude%s' % (str(resolution).replace('.', ''))` )

        Returns
        -------
        str
            longitude name
        """
        str_resolution = str(resolution).replace('.', '')
        str_resolution += '0' * (3 - len(str_resolution))  # Add 0 to have a str of 3 characters
        name = f"latitude{str_resolution}"
        if name in self.dataset.dims:
            return name
        else:
            raise ValueError(f"{name} wasn't found in the dataset. Please verify the resolution is correct")

    @property
    def time_name(self):
        """
        Get the name of the time variable in the dataset

        Returns
        -------
        str
            time name
        """
        return 'time'

    @property
    def acquisition_type(self):
        """
        Gives the acquisition type (swath, truncated_swath,daily_regular_grid, model_regular_grid)

        Returns
        -------
        str
            acquisition type

        """
        return 'model_regular_grid'

    @property
    def dataset(self):
        """
        Getter for the acquisition dataset

        Returns
        -------
        xarray.Dataset
            Acquisition dataset
        """
        return self._dataset

    @dataset.setter
    def dataset(self, value):
        """
        Setter of attribute `self.dataset`

        Parameters
        ----------
        value: xarray.Dataset
            new Dataset
        """
        self._dataset = value

    @property
    def orbit_segment_name(self):
        """
        Gives the name of the variable for orbit segmentation in dataset (Ascending / Descending). If value is None,
        so the orbit hasn't orbited segmentation

        Returns
        -------
        str | None
            Orbit segmentation variable name in the dataset. None if there isn't one.
        """
        return None

    @property
    def has_orbited_segmentation(self):
        """
        True if there is orbit segmentation in the dataset

        Returns
        -------
        bool
            Presence or not of an orbit segmentation
        """
        if self.orbit_segment_name is not None:
            return True
        else:
            return False

    @property
    def mission_name(self):
        """
        Name of the mission (or model)

        Returns
        -------
        str
            Mission name (ex: SMOS, S1, RS2, RCM, SMAP, HY2, ERA5)
        """
        return "ERA5"

    @property
    def wind_name(self):
        """
        Name of an important wind variable in the dataset

        Returns
        -------
        str
            Wind variable name

        """
        return 'v10'

    @property
    def longitude_name(self):
        """
        Get the name of the longitude variable in the dataset

        Returns
        -------
        str
            longitude name
        """
        if self._longitude_name is None:
            return ''
        else:
            return self._longitude_name

    @property
    def latitude_name(self):
        """
        Get the name of the latitude variable in the dataset

        Returns
        -------
        str
            longitude name
        """
        if self._latitude_name is None:
            return ''
        else:
            return self._latitude_name

    @longitude_name.setter
    def longitude_name(self, value):
        self._longitude_name = value

    @latitude_name.setter
    def latitude_name(self, value):
        self._latitude_name = value

    def reformat_meta(self):
        """
        Put both resolution of longitude and latitude at the same resolution. The resolution kept is the biggest one.
        """
        # ERA5 has a different format of latitude and longitude because it depends on the resolution
        longitude25 = self.longitude_name_res(0.25)
        latitude25 = self.latitude_name_res(0.25)
        longitude50 = self.longitude_name_res(0.50)
        latitude50 = self.latitude_name_res(0.50)
        ds = self.dataset.copy()
        # Adjust resolution of dimensions latitude025 and longitude025
        ds[latitude25] = ds[latitude50]
        ds[longitude25] = ds[longitude50]

        # Adjust resolution of variables variables that depend on latitude025 and longitude025
        variables_to_adjust = [var_name for var_name in ds.data_vars if
                               (longitude25 in ds[var_name].dims) and (latitude25 in ds[var_name].dims)]

        for var_name in variables_to_adjust:
            ds[var_name] = ds[var_name].interp(latitude025=ds[latitude25], longitude025=ds[longitude25])
        ds = ds.drop_vars([latitude25, longitude25])
        self.dataset = ds
        # New longitude and latitude names are these with the resolution of 050
        self.longitude_name = longitude50
        self.latitude_name = latitude50

    def rename_vars_in_coloc(self, dataset=None):
        """
        Rename variables from a dataset to homogenize the co-location product. If no dataset is explicit, so it is this
        of `self.dataset` which is used.

        Parameters
        ----------
        dataset: xarray.Dataset | None
            Dataset on which common vars must be renamed

        Returns
        -------
        xarray.Dataset
            Dataset with homogene variable names
        """
        if dataset is None:
            dataset = self.dataset
            # map the variable names in the dataset with the keys in common vars
        mapper = {
            self.wind_name: 'wind_direction',
            'u10': 'wind_speed',
        }
        for var in dataset.variables:
            if var in mapper.keys():
                key_in_common_vars = mapper[var]
                dataset = dataset.rename_vars({var: common_var_names[key_in_common_vars]})
        return dataset

    @property
    def unecessary_vars_in_coloc_product(self):
        """
        Get unecessary variables in co-location product

        Returns
        -------
        list[str]
            Unecessary variables in co-location product
        """
        return [self.time_name, 'land', 'nodata', 'ice', 'cloud']

    @property
    def necessary_attrs_in_coloc_product(self):
        """
        Get necessary dataset attributes in co-location product

        Returns
        -------
        list[str]
            Necessary dataset attributes in co-location product
        """
        # No attributes to the original dataset
        return []

    def rename_attrs_in_coloc_product(self, attr):
        """
        Get the new name of an attribute in co-location products from an original attribute

        Parameters
        ----------
        attr: str
            Attribute from the satellite dataset that needs to be renames for the co-location product.

        Returns
        -------
        str
            New attribute's name from the satellite dataset.
        """
        # No attributes to the original dataset
        return ""



import os
from .tools import call_sar_meta, open_l2, get_l2_footprint, extract_start_stop_dates_from_sar
from shapely.geometry import Polygon
import numpy as np


class GetSarMeta:
    def __init__(self, product_path, product_generation=False):
        self.product_path = product_path
        self.product_name = os.path.basename(self.product_path)
        self._l1_info = None
        self._l2_info = None
        self._time_name = None
        self._longitude_name = None
        self._latitude_name = None
        self.product_generation = product_generation
        if self.is_safe:
            self._l1_info = {'meta': call_sar_meta(self.product_path),
                             'dataset_names': [],
                             'submeta': {},
                             'times': {},
                             'footprints': {}
                             }
            self.fill_dataset_names()
            self.fill_submeta()
            self.fill_times()
            self.fill_footprints()
        else:
            self._time_name = 'time'
            self._longitude_name = 'lon'
            self._latitude_name = 'lat'
            self._l2_info = open_l2(product_path)

    def fill_submeta(self):
        """
        For a level 1 product, if it is a multi dataset, fills in a dictionary (`OpenSar._l1_info`) the metadata of
        sub-datasets
        """
        if self.is_safe:
            if self.multidataset:
                for ds_name in self._l1_info['dataset_names']:
                    self._l1_info['submeta'][ds_name] = call_sar_meta(ds_name)
            else:
                self._l1_info['submeta'][self._l1_info['dataset_names'][0]] = self._l1_info['meta']
        else:
            raise self.WrongProductTypeError("fill_submeta property only can be used for level 1 product")

    def fill_dataset_names(self):
        """
        For a level 1 product, if it is a multi dataset, fills in a dictionary (`OpenSar._l1_info`) the name of the
        sub-datasets
        """
        if self.is_safe:
            if (self.mission_name == 'S1') and self.multidataset:
                self._l1_info['dataset_names'] = [ds_name for ds_name in list(self._l1_info['meta']
                                                                              .subdatasets.index)]
            else:
                self._l1_info['dataset_names'] = [self.product_path]
        else:
            raise self.WrongProductTypeError("fill_dataset_names property only can be used for level 1 product")

    def fill_times(self):
        """
        For a level 1 product, if it is a multi dataset, fills in a dictionary (`OpenSar._l1_info`) the start/stop time
        of sub-datasets
        """
        if self.is_safe:
            for ds_name in self._l1_info['dataset_names']:
                tmp_dic = {
                    'start_date': self._l1_info['submeta'][ds_name].start_date,
                    'stop_date': self._l1_info['submeta'][ds_name].stop_date}
                self._l1_info['times'][ds_name] = tmp_dic
        else:
            raise self.WrongProductTypeError("fill_times property only can be used for level 1 product")

    def fill_footprints(self):
        """
        For a level 1 product, if it is a multi dataset, fills in a dictionary (`OpenSar._l1_info`) the footprint of
        sub-datasets
        """
        if self.is_safe:
            for ds_name in self._l1_info['dataset_names']:
                self._l1_info['footprints'][ds_name] = self._l1_info['submeta'][ds_name].footprint
        else:
            raise self.WrongProductTypeError("fill_footprints property only can be used for level 1 product")

    @property
    def multidataset(self):
        """
        Express if a product is a multi dataset or not.

        Returns
        -------
        bool
            Express if it is a multi dataset
        """
        if self.mission_name == 'S1':
            return self._l1_info['meta'].multidataset
        else:
            return False

    def datatree(self, ds_name):
        """
        For a level 1 product, getter for the datatree located in the metadata. Contains the main useful information

        Parameters
        ----------
        ds_name: str
            dataset_name (look into `OpenSar._l1_info['dataset_names']`) for available ones.

        Returns
        -------
        datatree.DataTree
            Main metadata information

        See Also
        --------
        `OpenSar._l1_info['dataset_names']`

        """
        if self.is_safe:
            return self._l1_info['submeta'][ds_name].dt
        else:
            raise self.WrongProductTypeError("datatree property only can be used for level 1 product")

    @property
    def mission_name(self):
        """
        From the product_name, get the mission name (ex : RADARSAT-2, RCM, SENTINEL-1)

        Returns
        -------
        str
            Mission name

        See Also
        --------
        `GetSarMeta.product_name`
        """
        if 'RS2' in self.product_name.upper():
            return 'RADARSAT-2'
        elif 'RCM' in self.product_name.upper():
            return 'RCM'
        elif 'S1A' in self.product_name.upper():
            return 'SENTINEL-1 A'
        elif 'S1B' in self.product_name.upper():
            return 'SENTINEL-1 B'
        else:
            raise TypeError("Unrecognized satellite name from %s" % str(self.product_name))

    @property
    def unecessary_vars_in_coloc_product(self):
        """
        Get unecessary variables in co-location product

        Returns
        -------
        list[str]
            Unecessary variables in co-location product
        """
        return []

    @property
    def necessary_attrs_in_coloc_product(self):
        """
        Get necessary dataset attributes in co-location product

        Returns
        -------
        list[str]
            Necessary dataset attributes in co-location product
        """
        return ['Conventions',
                'title',
                'institution',
                'reference',
                'measurementDate',
                'sourceProduct',
                'missionName',
                'polarization',
                'footprint',
                'l2ProcessingUtcTime',
                'version',
                'grid_mapping']

    def rename_attrs_in_coloc_product(self, attr):
        """
        Get the new name of an attribute in co-location products from an original attribute

        Parameters
        ----------
        attr: str
            Attribute from the satellite dataset that needs to be renames for the co-location product.

        Returns
        -------
        str
            New attribute's name from the satellite dataset.
        """
        # no attributes to rename
        mapper = {}
        if attr in mapper.keys():
            return mapper[attr]
        else:
            return attr

    @property
    def footprint(self):
        """
        Get footprint of the product

        Returns
        -------
        shapely.geometry.polygon.Polygon
            Footprint Polygon
        """
        if self.is_safe:
            entire_poly = Polygon()
            for ds_name in self._l1_info['dataset_names']:
                """sub_geoloc = self._l1_info['submeta'][ds_name].geoloc  # geoloc of a subdataset
                sub_geoloc = sub_geoloc.where((sub_geoloc.azimuthTime >= start_date) & \
                                              (sub_geoloc.azimuthTime <= stop_date), drop=True)
                stacked_subgeo = sub_geoloc.stack({"cells": ["line", "sample"]}).dropna(dim="cells")
                fp = MultiPoint(np.column_stack((stacked_subgeo.longitude, stacked_subgeo.latitude))).convex_hull"""
                self._l1_info['footprints'][ds_name] = self._l1_info['submeta'][ds_name].footprint
                fp = self._l1_info['footprints'][ds_name]
                entire_poly = entire_poly.union(fp)
            return entire_poly
        else:
            return get_l2_footprint(self._l2_info)

    @property
    def start_date(self):
        """
        Start acquisition date

        Returns
        -------
        numpy.datetime64
            Start date
        """
        if self.is_safe:
            start_dates = [np.datetime64(value['start_date']) for value in self._l1_info['times'].values()]
            return min(start_dates)
        else:
            # return np.datetime64(self._l2_info.attrs['firstMeasurementTime'])
            return extract_start_stop_dates_from_sar(self.product_path)[0]

    @property
    def stop_date(self):
        """
        Stop acquisition date

        Returns
        -------
        numpy.datetime64
            Stop date
        """
        if self.is_safe:
            stop_dates = [np.datetime64(value['stop_date']) for value in self._l1_info['times'].values()]
            return min(stop_dates)
        else:
            # return np.datetime64(self._l2_info.attrs['lastMeasurementTime'])
            return extract_start_stop_dates_from_sar(self.product_path)[1]

    @property
    def is_safe(self):
        """
        Know if a product is a Level 1 or Level 2. True if Level one

        Returns
        -------
        bool
            True if SAR product is a level 1

        """
        if self.product_name.endswith('.nc'):
            return False
        else:
            return True

    @property
    def acquisition_type(self):
        """
        Gives the acquisition type (swath, truncated_swath,daily_regular_grid, model_regular_grid)

        Returns
        -------
        str
            acquisition type

        """
        return 'truncated_swath'

    @property
    def dataset(self):
        """
        Getter for SAR dataset.
        NOTE: A SAR can be a L2 or a L1. This getter will be used in intersection functions. The choice has been made to
        use L1 only for listings (so we only need the footprint), and use L2 for co-location product. The dataset is
        needed only to create co-location product, so it is an alias of `self._l2_info`.

        Returns
        -------
        xarray.Dataset
            L2 SAR dataset
        """
        if self.is_safe:
            raise self.WrongProductTypeError('`dataset` property can only be used for level 1 products')
        else:
            return self._l2_info

    @property
    def longitude_name(self):
        """
        Get the name of the longitude variable in the dataset

        Returns
        -------
        str
            longitude name
        """
        if self.is_safe:
            raise NotImplementedError('`longitude_name` not implemented for safe products')
        else:
            return self._longitude_name

    @property
    def latitude_name(self):
        """
        Get the name of the latitude variable in the dataset

        Returns
        -------
        str
            latitude name
        """
        if self.is_safe:
            raise NotImplementedError('`latitude_name` not implemented for safe products')
        else:
            return self._latitude_name

    @property
    def time_name(self):
        """
        Get the name of the time variable in the dataset

        Returns
        -------
        str
            time name
        """
        if self.is_safe:
            raise NotImplementedError('`time_name` not implemented for safe products')
        else:
            return self._time_name

    @property
    def wind_name(self):
        """
        Name of an important wind variable in the dataset

        Returns
        -------
        str
            Wind variable name

        """
        return 'wind_speed'

    @property
    def orbit_segment_name(self):
        """
        Gives the name of the variable for orbit segmentation in dataset (Ascending / Descending). If value is None,
        so the orbit hasn't orbited segmentation

        Returns
        -------
        str | None
            Orbit segmentation variable name in the dataset. None if there isn't one.
        """
        return None

    @property
    def has_orbited_segmentation(self):
        """
        True if there is orbit segmentation in the dataset

        Returns
        -------
        bool
            Presence or not of an orbit segmentation
        """
        if self.orbit_segment_name is not None:
            return True
        else:
            return False

    class WrongProductTypeError(Exception):
        """
        Used for raising Exceptions when a function / property is called whereas it wasn't created for the specified
        type (Level 2 / Level 1)
        """
        pass

    @longitude_name.setter
    def longitude_name(self, value):
        self._longitude_name = value

    @latitude_name.setter
    def latitude_name(self, value):
        self._latitude_name = value

    @dataset.setter
    def dataset(self, value):
        if self.is_safe:
            raise self.WrongProductTypeError('`dataset` property can only be used for level 1 products')
        else:
            self._l2_info = value



from .tools import open_nc, correct_dataset
import os
import numpy as np


def extract_wind_speed(smos_dataset):
    return smos_dataset.where((np.isfinite(smos_dataset.wind_dir)), drop=True)


class GetHy2Meta:
    def __init__(self, product_path, product_generation=False):
        self.product_path = product_path
        self.product_name = os.path.basename(self.product_path)
        self.product_generation = product_generation
        self._time_name = 'time'
        self._longitude_name = 'lon'
        self._latitude_name = 'lat'
        self._dataset = open_nc(product_path).load()
        self.dataset = correct_dataset(self.dataset, self.longitude_name)

    @property
    def start_date(self):
        """
        Start acquisition time

        Returns
        -------
        numpy.datetime64
            Start time
        """
        return min(np.unique(self.dataset[self.time_name]))

    @property
    def stop_date(self):
        """
        Stop acquisition time

        Returns
        -------
        numpy.datetime64
            Stop time
        """
        return max(np.unique(self.dataset[self.time_name]))

    @property
    def longitude_name(self):
        """
        Get the name of the longitude variable in the dataset

        Returns
        -------
        str
            longitude name
        """
        return self._longitude_name

    @property
    def latitude_name(self):
        """
        Get the name of the latitude variable in the dataset

        Returns
        -------
        str
            latitude name
        """
        return self._latitude_name

    @property
    def time_name(self):
        """
        Get the name of the time variable in the dataset

        Returns
        -------
        str
            time name
        """
        return self._time_name

    @property
    def mission_name(self):
        """
        Get the mission name (ex : RADARSAT-2, RCM, SENTINEL-1, SMOS, SMAP,...)

        Returns
        -------
        str
            Mission name
        """
        return 'HY2'

    @property
    def acquisition_type(self):
        """
        Gives the acquisition type (swath, truncated_swath,daily_regular_grid, model_regular_grid)

        Returns
        -------
        str
            acquisition type

        """
        return 'swath'

    @property
    def dataset(self):
        """
        Getter for the acquisition dataset

        Returns
        -------
        xarray.Dataset
            Acquisition dataset
        """
        return self._dataset

    @dataset.setter
    def dataset(self, value):
        """
        Setter of attribute `self.dataset`

        Parameters
        ----------
        value: xarray.Dataset
            new Dataset
        """
        self._dataset = value

    @property
    def orbit_segment_name(self):
        """
        Gives the name of the variable for orbit segmentation in dataset (Ascending / Descending). If value is None,
        so the orbit hasn't orbited segmentation

        Returns
        -------
        str | None
            Orbit segmentation variable name in the dataset. None if there isn't one.
        """
        return None

    @property
    def has_orbited_segmentation(self):
        """
        True if there is orbit segmentation in the dataset

        Returns
        -------
        bool
            Presence or not of an orbit segmentation
        """
        if self.orbit_segment_name is not None:
            return True
        else:
            return False

    @property
    def mission_name(self):
        """
        Name of the mission (or model)

        Returns
        -------
        str
            Mission name (ex: SMOS, S1, RS2, RCM, SMAP, HY2, ERA5)
        """
        return "HY2"

    @longitude_name.setter
    def longitude_name(self, value):
        self._longitude_name = value

    @latitude_name.setter
    def latitude_name(self, value):
        self._latitude_name = value




"""Top-level package for coloc_sat."""

__author__ = """Yann Reynaud"""
__email__ = 'yann.reynaud.2@ifremer.fr'
__all__ = ['GetSarMeta', 'GetEra5Meta', 'GetHy2Meta', 'GetSmosMeta', 'GetWindSatMeta',
           'GetSmapMeta', 'ProductIntersection', 'GenerateColoc']

from .version import __version__
from .sar_meta import GetSarMeta
from .era5_meta import GetEra5Meta
from .hy2_meta import GetHy2Meta
from .smos_meta import GetSmosMeta
from .windsat_meta import GetWindSatMeta
from .smap_meta import GetSmapMeta
from .intersection import ProductIntersection
from .generate_coloc import GenerateColoc
from . import *




#	This routine reads version-7.0.1 RSS WindSat daily files 
#
# 	filename  name of data file with path in form satname_yyyymmdd_v7.0.1.gz
#	 where satname  = name of satellite (wsat)
#	       yyyy		= year
#		   mm		= month
#		   dd		= day of month
#   missing = fill value used for missing data;
#                  if None, then fill with byte codes (251-255)

##	the output values correspond to:
#	time		time of measurement in fractional hours GMT
#	sst			sea surface temperature in deg Celsius
#	windLF		10m surface wind, low frequency, in meters/second
#	windMF		10m surface wind, medium frequency, in meters/second
#	vapor		columnar or integrated water vapor in millimeters
#	cloud		cloud liquid water in millimeters
#	rain		rain rate in millimeters/hour
#   longitude	Grid Cell Center Longitude', LON = 0.25*x_grid_location - 0.125 degrees east
#   latitude	Grid Cell Center Latitude',  LAT = 0.25*y_grid_location - 90.125
#   land		Is this land?
#   ice			Is this ice?
#   nodata		Is there no data
#
#   if you need help with this python read code, contact support@remss.com


from .bytemaps import sys
from .bytemaps import Dataset
from .bytemaps import Verify
import xarray as xr


def to_xarray_dataset(raw):
    """
    Convert a WindSat dataset (format given by REMSS reader) to a xarray dataset.
    Parameters:
        raw (windsat_daily_v7.WindSatDaily): WindSat dataset in initial format.
    Returns
        (xarray.Dataset): WindSat dataset in xarray format.
    """

    def to_tuple(var):
        dims = var.coordinates
        data = var  # the Variable class is a subclass of ndarray
        attrs = {name: getattr(var, name) for name in raw._attributes() if name not in ["coordinates"]}
        return dims, data, attrs

    all_vars = {
        name: to_tuple(var)
        for name, var in raw.variables.items()
    }
    coordinate_names = raw._coordinates()
    data_vars = {k: v for k, v in all_vars.items() if k not in coordinate_names}
    coords = {k: v for k, v in all_vars.items() if k in coordinate_names}
    if "orbit_segment" not in coords:
        coords["orbit_segment"] = ("orbit_segment", ["descending", "ascending"])

    ds = xr.Dataset(coords=coords, data_vars=data_vars)
    ds.encoding["source"] = raw.filename

    return ds


class WindSatDaily(Dataset):
    """ Read daily WindSat bytemaps. """
    """
    Public data:
        filename = name of data file
        missing = fill value used for missing data;
                  if None, then fill with byte codes (251-255)
        dimensions = dictionary of dimensions for each coordinate
        variables = dictionary of data for each variable
    """

    def __init__(self, filename, missing=-999.):
        """
        Required arguments:
            filename = name of data file to be read (string)
                
        Optional arguments:
            missing = fill value for missing data,
                      default is the value used in verify file
        """
        self.filename = filename
        self.missing = missing
        Dataset.__init__(self)

    # Dataset:

    def _attributes(self):
        return ['coordinates','long_name','units','valid_min','valid_max']

    def _coordinates(self):
        return ('orbit_segment','variable','latitude','longitude')

    def _shape(self):
        return (2,9,720,1440)

    def _variables(self):
        return ['mingmt','sst','w-lf','w-mf','vapor','cloud','rain',
                'w-aw','wdir','longitude','latitude','land','ice','nodata']                

    # _default_get():
    
    def _get_index(self,var):
        return {'mingmt' : 0,
                'sst' : 1,
                'w-lf' : 2,
                'w-mf' : 3,
                'vapor' : 4,
                'cloud' : 5,
                'rain' : 6,
                'w-aw' : 7,
                'wdir' : 8,
                }[var]

    def _get_scale(self,var):
        return {'mingmt' : 6.0,
                'sst' : 0.15,
                'w-lf' : 0.2,
                'w-mf' : 0.2,
                'vapor' : 0.3,
                'cloud' : 0.01,
                'rain' : 0.1,
                'w-aw' : 0.2,
                'wdir' : 1.5,
                }[var]

    def _get_offset(self,var):
        return {'sst' : -3.0,
                'cloud' : -0.05,
                }[var]

    # _get_ attributes:

    def _get_long_name(self,var):
        return {'mingmt' : 'Time of Day UTC',
                'sst' : 'Sea Surface Temperature',
                'w-lf' : '10-m Surface Wind Speed (low frequency)',
                'w-mf' : '10-m Surface Wind Speed (medium frequency)',
                'vapor' : 'Columnar Water Vapor',
                'cloud' : 'Cloud Liquid Water',
                'rain' : 'Surface Rain Rate',
                'w-aw' : 'All-Weather 10-m Surface Wind Speed',
                'wdir' : 'Surface Wind Direction',
                'longitude' : 'Grid Cell Center Longitude',
                'latitude' : 'Grid Cell Center Latitude',
                'land' : 'Is this land?',
                'ice' : 'Is this ice?',
                'nodata' : 'Is there no data?',
                }[var]

    def _get_units(self,var):
        return {'mingmt' : 'Minutes since midnight GMT',
                'sst' : 'deg Celsius',
                'w-lf' : 'm/s',
                'w-mf' : 'm/s',
                'vapor' : 'mm',
                'cloud' : 'mm',
                'rain' : 'mm/hr',
                'w-aw' : 'm/s',
                'wdir' : 'deg oceanographic',
                'longitude' : 'degrees east',
                'latitude' : 'degrees north',
                'land' : 'True or False',
                'ice' : 'True or False',
                'nodata' : 'True or False',
                }[var]

    def _get_valid_min(self,var):
        return {'mingmt' : 0.0,
                'sst' : -3.0,
                'w-lf' : 0.0,
                'w-mf' : 0.0,
                'vapor' : 0.0,
                'cloud' : -0.05,
                'rain' : 0.0,
                'w-aw' : 0.2,
                'wdir' : 0.0,
                'longitude' : 0.0,
                'latitude' : -90.0,
                'land' : False,
                'ice' : False,
                'nodata' : False,
                }[var]

    def _get_valid_max(self,var):
        return {'mingmt' : 1440.0,
                'sst' : 34.5,
                'w-lf' : 50.0,
                'w-mf' : 50.0,
                'vapor' : 75.0,
                'cloud' : 2.45,
                'rain' : 25.0,
                'w-aw' : 50.0,
                'wdir' : 360.0,
                'longitude' : 360.0,
                'latitude' : 90.0,
                'land' : True,
                'ice' : True,
                'nodata' : True,
                }[var]


class DailyVerify(Verify):
    """ Contains info for verification. """
    
    def __init__(self,dataset):
        self.filename = 'windsat_V7_verify.txt'
        self.ilon1 = 170
        self.ilon2 = 175
        self.ilat1 = 274
        self.ilat2 = 278                
        self.iasc = 1
        self.variables = ['mingmt','sst','w-lf','w-mf','vapor',
                          'cloud','rain','w-aw','wdir']        
        self.startline = 97
        self.columns = {'mingmt' : 3,
                        'sst' : 4,
                        'w-lf' : 5,
                        'w-mf' : 6,
                        'vapor' : 7,
                        'cloud' : 8,
                        'rain' : 9,
                        'w-aw' : 10,
                        'wdir' : 11 }
        Verify.__init__(self,dataset)
        

if __name__ == '__main__':
    """ Automated testing. """

    # read daily:    
    wsat = WindSatDaily('wsat_20060409v7.0.1.gz')
    if not wsat.variables: sys.exit('file not found')
    
    # verify daily:
    verify = DailyVerify(wsat)
    if verify.success: print('successful verification for daily')
    else: sys.exit('verification failed for daily')
    print('')

    print('all tests completed successfully')
    print('')



""" Module for reading and verifying RSS gridded binary data files. """

import copy
import decimal
import gzip
import numpy as np
import sys
from collections import namedtuple
from collections import OrderedDict
from operator import mul
from functools import reduce


class Dataset:
    """ Base class for bytemap datasets. """
    """
    Public data:
        filename = name of data file
        missing = fill value used for missing data;
                  if None, then fill with byte codes (251-255)
        dimensions = dictionary of dimensions for each coordinate
        variables = dictionary of data for each variable
        
    All classes derived from Dataset must implement the following:
        _attributes() = returns list of attributes for each variable (list)
        _coordinates() = returns coordinates (tuple)
        _shape() = returns shape of raw data (tuple)
        _variables() = returns list of all variables to get (list)

    The derived class must provide "_get_" methods for the attributes.

    If the derived class provides "_get_" methods for the variables,
    those methods receive first priority.

    The "_get_" methods in this module receive second priority.

    The last priority is "_default_get", which requires:
        _get_index(var) = returns bmap index for var
        _get_scale(var) = returns bmap scale for var
        _get_offset(var) = returns bmap offset for var
    """

    def __init__(self):
        self.dimensions = self._get_dimensions()
        self.variables = self._get_variables()

    def _default_get(self, var, bmap):
        data = get_data(bmap, self._get_index(var))
        acopy = copy.deepcopy(data)
        bad = is_bad(data)
        try:
            data *= self._get_scale(var)
        except _NoValueFound:
            pass
        try:
            data += self._get_offset(var)
        except _NoValueFound:
            pass
        if self.missing == None:
            data[bad] = acopy[bad]
        else:
            data[bad] = self.missing
        return data

    def _dtype(self):
        return np.uint8

    def _get(self, var):
        try:
            return _get_(var, _from_=self)
        except _NoMethodFound:
            pass
        try:
            return _get_(var, _from_=thismodule())
        except _NoMethodFound:
            pass
        return self._default_get

    def _get_avariable(self, var, data):
        variable = self._get(var)(var, data)
        return Variable(var, variable, self)

    def _get_coordinates(self, var=None):
        if not var: return self._coordinates()
        if var in self._coordinates(): return (var,)
        return tuple([c for c in self._coordinates() if c != 'variable'])

    def _get_dimensions(self):
        dims = OrderedDict(zip(self._coordinates(), self._shape()))
        del dims['variable']
        return dims

    def _get_variables(self):
        data = OrderedDict()
        try:
            stream = readgz(self.filename)
        except:
            return data
        bmap = unpack(stream, shape=self._shape(), dtype=self._dtype())
        for var in self._variables():
            data[var] = self._get_avariable(var, bmap)
        return data


def readgz(filename):
    f = gzip.open(filename, 'rb')
    stream = f.read()
    f.close()
    return stream


def thismodule(): return sys.modules[__name__]


def unpack(stream, shape, dtype):
    count = reduce(mul, shape)
    return np.fromstring(stream, dtype=dtype, count=count).reshape(shape)


""" Library of Methods for _get_ Functions: """


def btest(ival, ipos):
    """Same usage as Fortran btest function."""
    return (ival & (1 << ipos)) != 0


def cosd(x): return np.cos(np.radians(x))


def get_data(bmap, indx, dtype=np.float64):
    """Return numpy array of dytpe for variable in bmap given by indx."""
    return np.array(np.squeeze(bmap[..., indx, :, :]), dtype=dtype)


def get_uv(speed, direction):
    """
    Given speed and direction (degrees oceanographic),
    return u (zonal) and v (meridional) components.
    """
    u = speed * sind(direction)
    v = speed * cosd(direction)
    return u, v


def ibits(ival, ipos, ilen):
    """Same usage as Fortran ibits function."""
    ones = ((1 << ilen) - 1)
    return (ival & (ones << ipos)) >> ipos


def is_bad(bmap, maxvalid=250):
    """Return mask where data are bad."""
    return bmap > maxvalid


def sind(x): return np.sin(np.radians(x))


where = np.where

""" Library of Named Exceptions: """

_NoMethodFound = AttributeError

_NoValueFound = (AttributeError, KeyError)

_NotFound = AttributeError

""" Library of Named _get_ Functions: """


def _get_(var, _from_): return getattr(_from_, '_get_' + var)


def _get_ice(var, bmap, indx=0, icevalue=252):
    return get_data(bmap, indx, dtype=bmap.dtype) == icevalue


def _get_land(var, bmap, indx=0, landvalue=255):
    return get_data(bmap, indx, dtype=bmap.dtype) == landvalue


def _get_latitude(var, bmap, nlat=720, dlat=0.25, lat0=-89.875):
    if np.shape(bmap)[-2] != nlat: sys.exit('Latitude mismatch')
    return np.array([dlat * ilat + lat0 for ilat in range(nlat)])


def _get_longitude(var, bmap, nlon=1440, dlon=0.25, lon0=0.125):
    if np.shape(bmap)[-1] != nlon: sys.exit('Longitude mismatch')
    return np.array([dlon * ilon + lon0 for ilon in range(nlon)])


def _get_nodata(var, bmap, indx=0):
    return is_bad(get_data(bmap, indx, dtype=bmap.dtype))


class Variable(np.ndarray):
    """ Variable exists solely to subclass numpy array with attributes. """

    def __new__(cls, var, data, dataset):
        obj = np.asarray(data).view(cls)
        for attr in dataset._attributes():
            get = _get_(attr, _from_=dataset)
            setattr(obj, attr, get(var))
        return obj


class Verify:
    """ Base class for bytemap read verification. """
    """
    Public data:
        data = OrderedDict of OneOb-namedtuple lists for each variable
        success = True/False
    
    The derived class must supply the following:

    For all files:
        filename = name of verify file
        variables = list of variables to verify
        
    The following indices (1-based):
        ilon1 = longitude index
        ilon2 = longitude index
        ilat1 = latitude index
        ilat2 = latitude index
        iasc = asc/dsc index (daily only)

    For files organized as a list:
        startline = starting line number of data (integer)
        columns = column numbers for each variable (dictionary)

    For files organized as arrays:
        startline = starting line number of data for each variable (dict)

    The startline and columns are counting starting from 1.
    """

    def __init__(self, dataset):
        self._file = [tokenize(line) for line in readtext(self.filename)]
        self.data = self._get_data()
        self.success = verify(dataset, self)

    def _asc(self):
        try:
            return zerobased(self.iasc)
        except _NotFound:
            return Ellipsis

    def _get_avariable(self, var):
        data = []
        indices = np.ndindex(self._nlat(), self._nlon())
        for ilat, ilon in indices:
            data.append(self._get_oneob(var, ilon, ilat))
        return data

    def _get_data(self):
        data = OrderedDict()
        for var in self.variables:
            data[var] = self._get_avariable(var)
        return data

    def _get_line_word(self, var, ilon, ilat):
        if self._islist():
            return self._get_line_word_list(var, ilon, ilat)
        else:
            return self._get_line_word_array(var, ilon, ilat)

    def _get_line_word_array(self, var, ilon, ilat):
        iline = zerobased(self.startline[var]) + ilat
        iword = ilon
        return iline, iword

    def _get_line_word_list(self, var, ilon, ilat):
        iline = zerobased(self.startline) + ilat * self._nlon() + ilon
        iword = zerobased(self.columns[var])
        return iline, iword

    def _get_oneob(self, var, ilon, ilat):
        iline, iword = self._get_line_word(var, ilon, ilat)
        avalue = self._file[iline][iword]
        return OneOb(self._lon(ilon), self._lat(ilat), self._asc(),
                     float(avalue), places(avalue))

    def _islist(self):
        return hasattr(self, 'columns')

    def _lat(self, ilat):
        return zerobased(self.ilat1) + ilat

    def _lon(self, ilon):
        return zerobased(self.ilon1) + ilon

    def _nlat(self):
        return self.ilat2 - self.ilat1 + 1

    def _nlon(self):
        return self.ilon2 - self.ilon1 + 1


OneOb = namedtuple('OneOb', 'lon lat asc val ndp')
"""
OneOb corresponds to one observation from verify file with:
    lon = longitude index
    lat = latitude index
    asc = ascending/descending index
    val = verify value
    ndp = number of decimal places given in verify
    The (asc,lat,lon) indices are 0-based.
"""


def places(astring):
    """
    Given a string representing a floating-point number,
    return number of decimal places of precision (note: is negative).
    """
    return decimal.Decimal(astring).as_tuple().exponent


def readtext(filename):
    f = open(filename, 'r')
    lines = f.readlines()
    f.close()
    return lines


def tokenize(line): return [item.strip() for item in line.split()]


def verify(dataset, verify):
    """ Verify data were read correctly. """
    """
    Required arguments:
        dataset = a read Dataset instance
        verify = a Verify instance
    Returns:
        success = True or False
    """
    success = True
    for var in verify.variables:
        for ob in verify.data[var]:
            readval = dataset.variables[var][ob.asc, ob.lat, ob.lon]
            diff = abs(ob.val - readval)
            match = diff < pow(10, ob.ndp)
            if not match: success = False
            print(' '.join([str(ob.lon), str(ob.lat), str(var),
                            str(ob.val), str(readval), str(diff), str(match)]))
    return success


def zerobased(indx): return indx - 1


if __name__ == '__main__':
    link = 'http://www.remss.com/terms_of_data_use/terms_of_data_use.html'
    print('Remote Sensing Systems')
    print('444 Tenth Street, Suite 200')
    print('Santa Rosa, CA 95401, USA')
    print('FTP: ftp://ftp.ssmi.com')
    print('Web: http://www.remss.com')
    print('Support: support@remss.com')
    print('Terms of Data Use: ' + link)



import os
import glob
from pathlib import Path

import xarray as xr
import yaml
from shapely import wkt
from shapely.geometry import Polygon
import numpy as np
import fsspec
from datetime import datetime
from xsar.raster_readers import resource_strftime
import re

param_config = None


def get_config_path():
    if param_config is not None:
        return param_config
    else:
        # determine the config file we will use (config.yml by default, and a local config if one is present)
        local_config_pontential_path = Path(os.path.join('~', 'coloc_sat', 'localconfig.yml')).expanduser()
        if local_config_pontential_path.exists():
            config_path = local_config_pontential_path
        else:
            config_path = Path(os.path.join(os.path.dirname(__file__), 'config.yml'))
        return config_path


def load_config():
    with open(get_config_path(), 'r') as file:
        config = yaml.safe_load(file)
    return config


common_var_names = load_config().get('common_var_names', {})


def set_config(config_path: str):
    global param_config
    global common_var_names

    param_config = config_path
    common_var_names = load_config().get('common_var_names', {})


def get_acquisition_root_paths(ds_name):
    paths_dict = load_config().get('paths', {})
    return paths_dict[ds_name]


def call_meta_class(file, product_generation=False):
    sar_satellites = ['RS2', 'S1A', 'S1B', 'RCM1', 'RCM2', 'RCM3']
    basename = os.path.basename(file).upper()
    if basename.split('_')[0].split('-')[0] in sar_satellites:
        from .sar_meta import GetSarMeta
        return GetSarMeta(file, product_generation=product_generation)
    elif basename.startswith('SM_'):
        from .smos_meta import GetSmosMeta
        return GetSmosMeta(file, product_generation=product_generation)
    elif basename.startswith('WSAT_'):
        from .windsat_meta import GetWindSatMeta
        return GetWindSatMeta(file, product_generation=product_generation)
    elif basename.split('_')[1] == 'SMAP':
        from .smap_meta import GetSmapMeta
        return GetSmapMeta(file, product_generation=product_generation)
    elif basename.split('_')[3] == 'HY':
        from .hy2_meta import GetHy2Meta
        return GetHy2Meta(file, product_generation=product_generation)
    elif basename.startswith('ERA_5'):
        from .era5_meta import GetEra5Meta
        return GetEra5Meta(file, product_generation=product_generation)
    else:
        raise ValueError(f"Can't recognize satellite type from product {basename}")


def get_all_comparison_files(start_date=None, stop_date=None, ds_name='SMOS', input_ds=None, level=None):
    """
    Return all existing product for a specific sensor (ex : SMOS, RS2, RCM, S1, HY2, ERA5)

    Parameters
    ----------
    start_date: numpy.datetime64 | None
        Start date for the research
    stop_date: numpy.datetime64 | None
        Stop date for the research
    ds_name: str
        Sensor name
    input_ds: None | list[str] | str
        If we don't want to make the research in all the files, but only in a subset,
        it can be explicit there as a list of product paths.
        It can also be explicit as a txt file path in which are located all the product paths
        Value is None if the research must be on all the files
    level: int | None
        When ds_name is SAR, precise the value of the product level. If it is None, get all SAR levels. Useless to give
        it a value when ds_name is something else then a SAR ('S1', 'RS2', 'RCM'). Values can be 1, 2 or None
        (default value).

    Returns
    -------
    List[str]
        Path of all existing products
    """

    def insert_date_and_day_of_year(str_expression, datetime_obj, day_of_year):
        """
        Convert special characters in an expression (like %m, %Y, %d and %(dayOfYear)) by the corresponding strings
        (extracted from the date) and day of year (if one is needed)

        Parameters
        ----------
        str_expression: str
            Expression that contains special characters.
            Example : '/home/ref-smoswind-public/data/v3.0/l3/data/reprocessing/%Y/%(dayOfYear)/*%Y%m%d*.nc'
        datetime_obj: datetime.datetime
            Date that need to be parsed in the expression
        day_of_year: str | int
            Day number of the year

        Returns
        -------
        str
            New expression with date and day of the year parsed.
        """
        str_expression = datetime_obj.strftime(str_expression)
        str_expression = str_expression.replace('%(dayOfYear)', day_of_year)
        return str_expression

    def research_files(expression):
        if (input_ds is not None) and isinstance(input_ds, list):
            return match_expression_in_list(expression=expression, str_list=input_ds)
        elif (input_ds is not None) and isinstance(input_ds, str):
            with open(input_ds, 'r') as file:
                lines = file.readlines()
            file.close()
            files_list = [line.strip() for line in lines]
            return match_expression_in_list(expression=expression, str_list=files_list)
        elif input_ds is None:
            return glob.glob(expression)
        else:
            raise ValueError('Type of input_ds must be a list or None')

    def get_last_generation_files(files_list):
        """
        From a list of SMOS paths, return only the paths with the latest generation

        Parameters
        ----------
        files_list: List[str]
            List of SMOS paths

        Returns
        -------
        List[str]
            Latest generation SMOS paths

        """

        def extract_smos_sort_keys(string):
            """
            From a SMOS path, extract the orbit (Ascending or Descending), the date and the generation number. It is
            used to sort a list of SMOS paths

            Parameters
            ----------
            string: str
                SMOS path

            Returns
            -------
            (str, int, int)
                Primary and secondary sort keys (orbit, date, generation number)
            """
            basename = os.path.basename(string)
            key1 = basename.split('_')[-5]
            key2 = int(basename.split('_')[-4])
            key3 = int(basename.split('_')[-2])
            return key1, key2, key3

        if len(files_list) == 0:
            return files_list
        else:
            final_files = []
            sorted_files = sorted(files_list, key=extract_smos_sort_keys)
            last_generation_file = sorted_files[0]
            for index, file in enumerate(sorted_files):
                # prefix is the same when only the generation is different
                prefix = '_'.join(os.path.basename(file).split('_')[:-2])
                if prefix == '_'.join(os.path.basename(last_generation_file).split('_')[:-2]):
                    # if the generation is greater, we increase the reference generation
                    if extract_smos_sort_keys(file)[2] >= extract_smos_sort_keys(last_generation_file)[2]:
                        last_generation_file = file
                else:
                    final_files.append(last_generation_file)
                    last_generation_file = file
                # The last files isn't added when it is a new product, so we add it
                if index == len(sorted_files) - 1:
                    final_files.append(file)
            return final_files

    map_levels = {
        1: 'L1',
        2: 'L2'
    }

    root_paths = get_acquisition_root_paths(ds_name)
    product_levels = []
    if level is not None:
        product_levels = [map_levels[level]]
    elif (ds_name == 'S1') or (ds_name == 'RS2') or (ds_name == 'RCM'):
        product_levels = list(root_paths.keys())
    files = []
    schemes = date_schemes(start_date, stop_date)
    if ds_name == 'SMOS':
        # get all netcdf files which contain the days in schemes
        for root_path in root_paths:
            for scheme in schemes:
                date = datetime.strptime(scheme, '%Y%m%d')
                parsed_path = insert_date_and_day_of_year(str_expression=root_path, datetime_obj=date,
                                                          day_of_year=schemes[scheme]['dayOfYear'])
                files += research_files(parsed_path)
        files = get_last_generation_files(files)
    elif ds_name == 'HY2':
        # get all netcdf files which contain the days in schemes
        for root_path in root_paths:
            for scheme in schemes:
                date = datetime.strptime(scheme, '%Y%m%d')
                parsed_path = insert_date_and_day_of_year(str_expression=root_path, datetime_obj=date,
                                                          day_of_year=schemes[scheme]['dayOfYear'])
                files += research_files(parsed_path)
        if (start_date is not None) and (stop_date is not None):
            # remove files for which hour doesn't correspond to the selected times
            for f in files.copy():
                start_hy, stop_hy = extract_start_stop_dates_from_hy(f)
                if (stop_hy < start_date) or (start_hy > stop_date):
                    files.remove(f)
    elif ds_name == 'S1':
        for lvl in product_levels:
            for root_path in root_paths[lvl]:
                for scheme in schemes:
                    date = datetime.strptime(scheme, '%Y%m%d')
                    parsed_path = insert_date_and_day_of_year(str_expression=root_path, datetime_obj=date,
                                                              day_of_year=schemes[scheme]['dayOfYear'])
                    files += research_files(parsed_path)
    elif ds_name == 'RS2':
        for lvl in product_levels:
            for root_path in root_paths[lvl]:
                for scheme in schemes:
                    date = datetime.strptime(scheme, '%Y%m%d')
                    parsed_path = insert_date_and_day_of_year(str_expression=root_path, datetime_obj=date,
                                                              day_of_year=schemes[scheme]['dayOfYear'])
                    files += research_files(parsed_path)
    elif ds_name == 'RCM':
        for lvl in product_levels:
            for root_path in root_paths[lvl]:
                for scheme in schemes:
                    date = datetime.strptime(scheme, '%Y%m%d')
                    parsed_path = insert_date_and_day_of_year(str_expression=root_path, datetime_obj=date,
                                                              day_of_year=schemes[scheme]['dayOfYear'])
                    files += research_files(parsed_path)
    elif ds_name == 'ERA5':
        for root_path in root_paths:
            if (start_date is not None) and (stop_date is not None):
                files = get_nearest_era5_files(start_date, stop_date, root_path)
            else:
                files = research_files(root_path.replace('%Y', '*').replace('%m', '*').replace('%d', '*'))
    elif ds_name == 'WS':
        for root_path in root_paths:
            for scheme in schemes:
                date = datetime.strptime(scheme, '%Y%m%d')
                parsed_path = insert_date_and_day_of_year(str_expression=root_path, datetime_obj=date,
                                                          day_of_year=schemes[scheme]['dayOfYear'])
                files += research_files(parsed_path)
    elif ds_name == 'SMAP':
        for root_path in root_paths:
            for scheme in schemes:
                date = datetime.strptime(scheme, '%Y%m%d')
                parsed_path = insert_date_and_day_of_year(str_expression=root_path, datetime_obj=date,
                                                          day_of_year=schemes[scheme]['dayOfYear'])
                files += research_files(parsed_path)
    if (start_date is not None) and (stop_date is not None):
        if ds_name in ['S1', 'RS2', 'RCM']:
            for f in files.copy():
                start, stop = extract_start_stop_dates_from_sar(f)
                if (stop < start_date) or (start > stop_date):
                    files.remove(f)
    return files


def match_expression_in_list(expression, str_list):
    regex_expr = re.sub(r'\*', r'.*', expression)
    return [path for path in str_list if re.match(regex_expr, path)]


def get_nearest_era5_files(start_date, stop_date, resource, step=1):
    """
    Get a list of era5 files

    Parameters
    ----------
    start_date: numpy.datetime64
        Start date for the research of era 5 files
    stop_date: numpy.datetime64
        End date for the research of era 5 files
    resource: str
        resource string, with strftime template
    step: int
        hour step between 2 files

    Returns
    -------
    list[str]
        Concerned ERA5 files
    """
    files = []
    date = start_date.astype('datetime64[ns]')
    while date < stop_date:
        datetime_date = datetime.utcfromtimestamp(date.astype(int) * 1e-9)
        closest_date, filename = resource_strftime(resource, step=step, date=datetime_date)
        if filename not in files:
            files.append(filename)
        date += np.timedelta64(step, 'm')
    return files


def cross_antemeridian(dataset):
    """True if footprint cross antemeridian"""
    return ((np.max(dataset.lon) - np.min(
        dataset.lon)) > 180).item()


def correct_dataset(dataset, lon_name='lon'):
    """
    Get acquisition dataset depending on latitude and longitude. Apply correction if needed when it crosses antemeridian.
    Longitude values are ranging between -180 and 180 degrees.

    Parameters
    ----------
    dataset: xarray.Dataset
        Acquisition dataset
    lon_name: str
        name of the longitude dimension in the dataset. `lon` by default.

    Returns
    -------
    xarray.Dataset
        Acquisition dataset depending on longitude and latitude.
    """

    def cross_antemeridian(ds):
        """True if footprint cross antemeridian"""
        return ((np.max(ds[lon_name]) - np.min(
            ds[lon_name])) > 180).item()

    lon = dataset[lon_name]
    if cross_antemeridian(dataset):
        lon = (lon + 180) % 360
    dataset = dataset.assign_coords(**{lon_name: lon - 180})
    if dataset[lon_name].ndim == 1:
        dataset = dataset.sortby(lon_name)
    return dataset


def date_schemes(start_date, stop_date):
    schemes = {}
    if (start_date is not None) and (stop_date is not None):
        date = np.datetime64(start_date, 's')
        while date.astype('datetime64[D]') <= stop_date.astype('datetime64[D]'):
            scheme = str(date.astype('datetime64[D]')).replace('-', '')
            year = str(date.astype('datetime64[Y]'))
            month = str(date.astype('datetime64[M]')).split('-')[1]
            day_of_year = date.astype(datetime).strftime('%j')
            date += np.timedelta64(1, 'D')
            tmp_dic = {'year': year,
                       'dayOfYear': day_of_year,
                       'month': month}
            schemes[scheme] = tmp_dic
    else:
        schemes = {
            '*': {
                'year': '*',
                'dayOfYear': '*',
                'month': '*'
            }
        }
    return schemes


def extract_start_stop_dates_from_hy(product_path):
    ds = open_nc(product_path)
    unique_time = np.unique(ds.time)
    return min(unique_time), max(unique_time)


def parse_date(date):
    """
    Parse a date at the format %Y%Y%Y%Y%M%M%D%D%H%H%M%M%S%S, to the format numpy.datetime64

    Parameters
    ----------
    date: str
        date at the format %Y%Y%Y%Y%M%M%D%D%H%H%M%M%S%S

    Returns
    -------
    numpy.datetime64
        parsed date
    """
    if not isinstance(date, str):
        raise ValueError('Argument date must be a string')
    if len(date) != 14:
        raise ValueError("Date isn't at the good format, please use the format %Y%Y%Y%Y%M%M%D%D%H%H%M%M%S%S")
    # formatted_date_string = f"{date[0:4]}-{date[4:6]}-{date[6:8]}T{date[8:10]}:{date[10:12]}:{date[12:16]}"
    return np.datetime64(datetime.strptime(date, '%Y%m%d%H%M%S'))


def extract_start_stop_dates_from_sar(product_path):
    """
    Get the start and stop date for a SAR product filename. Caution: Level 1 --> for RS2 and RCM products,
    filename only contains the start date, so stop date = start date + 5 minutes

    Parameters
    ----------
    product_path: str
        path of the product

    Returns
    -------
    np.datetime64, np.datetime64
        Tuple that contains the start and the stop dates
    """
    separators = {
        'L1': '_',
        'L2': '-'
    }
    # All level 2 products have a start and a stop date
    index_l2 = {
        'start': 4,
        'stop': 5
    }
    # All S1 level 1 have a start and a stop date
    index_l1_sentinel = {
        'start': -5,
        'stop': -4
    }
    # All RCM and RS2 level 1 only have a start date, divided in a date (%Y%Y%Y%Y%M%M%D%D) and a time (%H%H%M%M%S%S)
    index_l1_radarsat = {
        'date': 5,
        'time': 6
    }
    basename = os.path.basename(product_path)
    upper_basename = basename.upper()
    # level 2 products
    if basename.endswith('.nc'):
        split_basename = upper_basename.split(separators['L2'])
        str_start_date = split_basename[index_l2['start']].replace('T', '')
        str_stop_date = split_basename[index_l2['stop']].replace('T', '')
        start, stop = parse_date(str_start_date), parse_date(str_stop_date)
    # level 1 products
    else:
        split_basename = upper_basename.split(separators['L1'])
        # S1 products
        if upper_basename.startswith('S1'):
            str_start_date = split_basename[index_l1_sentinel['start']].replace('T', '')
            str_stop_date = split_basename[index_l1_sentinel['stop']].replace('T', '')
            start, stop = parse_date(str_start_date), parse_date(str_stop_date)
        elif upper_basename.startswith('RCM') or upper_basename.startswith('RS2'):
            str_start_date = split_basename[index_l1_radarsat['date']] + split_basename[index_l1_radarsat['time']]
            start = parse_date(str_start_date)
            # we only have the start date, so stop date = start date + 5 minutes
            stop = start + np.timedelta64(5, 'm')
        else:
            raise ValueError(f"Can't recognize if the product {basename} is a RCM, a S1 or a RS2")
    return start, stop


def call_sar_meta(dataset_id):
    """
    Call the appropriate metadata for a SAR Level 1 product depending on the dataset id.

    Parameters
    ----------
    dataset_id: str
        Path to the Level 1 product on which the metadata must be accessed

    Returns
    -------
    xsar.Sentinel1Meta | xsar.RadarSat2Meta | xsar.RcmMeta
        Object that contains the metadata
    """
    if isinstance(dataset_id, str) and "S1" in dataset_id:
        from xsar import Sentinel1Meta
        sar_meta = Sentinel1Meta(dataset_id)
    elif isinstance(dataset_id, str) and "RS2" in dataset_id:
        from xsar import RadarSat2Meta
        sar_meta = RadarSat2Meta(dataset_id)
    elif isinstance(dataset_id, str) and "RCM" in dataset_id:
        from xsar import RcmMeta
        sar_meta = RcmMeta(dataset_id)
    else:
        raise TypeError("Unknown dataset id type from %s" % str(dataset_id))
    return sar_meta


def find_l2_nc(product_path):
    if os.path.isdir(product_path):
        nc_product = glob.glob(os.path.join(product_path, '*owi*.nc'))
        if len(nc_product) > 1:
            raise ValueError(f"Many netcdf files can be read for this product, please select an only one in the " +
                             f"following list : {nc_product}")
        else:
            nc_product = nc_product[0]
    else:
        nc_product = product_path
    return nc_product


def open_l2(product_path):
    """
    Open a SAR level 2 product as a dataset

    Parameters
    ----------
    product_path: str
        Path to level 2 product that must be opened

    Returns
    -------
    xarray.Dataset
        Level 2 SAR product
    """
    nc_product = find_l2_nc(product_path)
    fs = fsspec.filesystem("file")
    return xr.open_dataset(fs.open(nc_product), engine='h5netcdf')


def convert_str_to_polygon(poly_str):
    """
    Convert a string to a shapely Polygon object.

    Parameters
    ----------
    poly_str: str
        string that represents a shapely Polygon object. Example :
        `POLYGON ((-95.07443 25.2053, -92.21184 25.696226, -92.74229 28.370426, -95.674324 27.886456, -95.07443 25.2053))`

    Returns
    -------
    shapely.geometry.polygon.Polygon
        Polygon
    """
    return wkt.loads(poly_str)


def get_l2_footprint(dataset):
    """
    Get footprint of a Level 2 SAR product

    Parameters
    ----------
    dataset: xarray.Dataset
        Dataset of the Level 2 product

    Returns
    -------
    shapely.geometry.polygon.Polygon
        Footprint of the product as a polygon
    """
    if 'footprint' in dataset.attrs:
        return convert_str_to_polygon(dataset.attrs['footprint'])
    else:
        footprint_dict = {}
        for ll in ['owiLon', 'owiLat']:
            footprint_dict[ll] = [
                dataset[ll].isel(owiAzSize=a, owiRaSize=x).values for a, x in [(0, 0), (0, -1), (-1, -1), (-1, 0)]
            ]
        corners = list(zip(footprint_dict['owiLon'], footprint_dict['owiLat']))
        return Polygon(corners)


def open_nc(product_path):
    """
    Open a netcdf file using `xarray.open_dataset`

    Parameters
    ----------
    product_path: str
        Absolute path to the netcdf

    Returns
    -------
    xarray.Dataset
        netcdf content
    """
    fs = fsspec.filesystem("file")
    return xr.open_dataset(fs.open(product_path))


def open_smos_file(product_path):
    """
    Open a smos file as a dataset

    Parameters
    ----------
    product_path: str
        Path to the smos product that must be opened

    Returns
    -------
    xarray.Dataset
        Smos product
    """
    fs = fsspec.filesystem("file")
    return xr.open_dataset(fs.open(product_path), engine='h5netcdf')


def convert_mingmt(meta_acquisition):
    """
    Convert a time array since midnight GMT format (from an acquisition dataset) to the numpy.datetime64 format.

    Parameters
    ----------
    meta_acquisition: GetSmapMeta | GetWindSatMeta
        Metadata class of the acquisition that must have it minute variable corrected

    Returns
    -------
    xarray.Dataset
        Co-located WindSat dataset with time reformatted. New time variable has the name specified in
        `meta_acquisition.time_name`
    """
    ds = meta_acquisition.dataset
    input_time = ds[meta_acquisition.minute_name]
    if (np.dtype(input_time) == np.dtype('float64')) or (np.dtype(input_time) == np.dtype(int)):
        input_time = input_time.astype('timedelta64[m]')
    ds[meta_acquisition.time_name] = np.array(meta_acquisition.day_date, dtype="datetime64[ns]") + input_time
    return ds.drop_vars([meta_acquisition.minute_name])


def extract_name_from_meta_class(obj):
    """
    Extract type of satellite (or name of a model).

    Parameters
    ----------
    obj: coloc_sat.GetSarMeta | coloc_sat.GetSmosMeta | coloc_sat.GetSmapMeta | coloc_sat.GetHy2Meta |
    coloc_sat.GetEra5Meta | coloc_sat.GetWindsatMeta
        Meta object

    Returns
    -------
    str
        Type of a satellite
    """
    class_name = obj.__class__.__name__
    pattern = r"Get(\w+)Meta"
    match = re.match(pattern, class_name)
    if match:
        return match.group(1)
    else:
        return None


def date_means(date1, date2):
    """
    Get means of 2 datetimes

    Parameters
    ----------
    date1: datetime.datetime
        Date 1
    date2: datetime.datetime
        date 2
    Returns
    -------
    datetime.datetime
        Means of 2 dates
    """
    diff = date2 - date1
    half_diff = diff / 2
    mean = date1 + half_diff

    return mean


def mean_time_diff(start1, stop1, start2, stop2):
    """
    Get mean of the time difference between 2 date ranges.

    Parameters
    ----------
    start1: datetime.datetime | str
        Start of the first range date.
    stop1: datetime.datetime | str
        Stop of the first range date.
    start2: datetime.datetime | str
        Start of the second range date.
    stop2: datetime.datetime | str
        Stop of the second range date.
    Returns
    -------
    datetime.datetime
        Mean of the time difference between 2 date ranges.
    """
    if isinstance(start1, str):
        start1 = np.datetime64(start1)
    if isinstance(stop1, str):
        stop1 = np.datetime64(stop1)
    if isinstance(start2, str):
        start2 = np.datetime64(start2)
    if isinstance(stop2, str):
        stop2 = np.datetime64(stop2)
    mean1 = date_means(start1, stop1)
    mean2 = date_means(start2, stop2)
    return abs(mean1 - mean2)


def reformat_meta(meta):
    """
    Rename the longitude and latitude names in the dataset and in the properties.

    Parameters
    ----------
    meta: coloc_sat.GetSarMeta | coloc_sat.GetSmosMeta | coloc_sat.GetSmapMeta | coloc_sat.GetHy2Meta |
    coloc_sat.GetEra5Meta | coloc_sat.GetWindsatMeta
        Meta object

    Returns
    -------
    coloc_sat.GetSarMeta | coloc_sat.GetSmosMeta | coloc_sat.GetSmapMeta | coloc_sat.GetHy2Meta |
    coloc_sat.GetEra5Meta | coloc_sat.GetWindsatMeta
        Meta object reformatted
    """
    satellite_type = extract_name_from_meta_class(meta)
    if satellite_type == 'Sar':
        if meta.is_safe:
            # Safe product don't have dataset property
            return meta
    ds = meta.dataset
    # rename longitude, latitude by references name and modify concerned attributes in the metaobjects
    if meta.longitude_name != common_var_names['longitude']:
        ds = ds.rename({meta.longitude_name: common_var_names['longitude']})
    if meta.latitude_name != common_var_names['latitude']:
        ds = ds.rename({meta.latitude_name: common_var_names['latitude']})
    meta.dataset = ds
    meta.longitude_name = common_var_names['longitude']
    meta.latitude_name = common_var_names['latitude']
    return meta



import os
import numpy as np
from datetime import datetime
import xarray as xr

from .tools import open_nc, convert_mingmt, correct_dataset, common_var_names


class GetSmapMeta:
    def __init__(self, product_path, product_generation=False):
        self.product_path = product_path
        self.product_name = os.path.basename(self.product_path)
        self.product_generation = product_generation
        self._time_name = 'time'
        self._longitude_name = 'lon'
        self._latitude_name = 'lat'
        self._dataset = open_nc(product_path).load()
        self.dataset = self.add_source_reference_attribute(ds=self.dataset)
        self.dataset = correct_dataset(self.dataset, self.longitude_name)
        self.dataset = convert_mingmt(self)
        # Modify orbit values by ascending and descending to be more significant
        self.dataset[self.orbit_segment_name] = \
            xr.where(self.dataset[self.orbit_segment_name] == 0, 'ascending', 'descending')

    @property
    def longitude_name(self):
        """
        Get the name of the longitude variable in the dataset

        Returns
        -------
        str
            longitude name
        """
        return self._longitude_name

    @property
    def latitude_name(self):
        """
        Get the name of the latitude variable in the dataset

        Returns
        -------
        str
            latitude name
        """
        return self._latitude_name

    @property
    def time_name(self):
        """
        Get the name of the time variable in the dataset

        Returns
        -------
        str
            time name
        """
        return self._time_name

    @property
    def dataset(self):
        """
        Getter for the acquisition dataset

        Returns
        -------
        xarray.Dataset
            Acquisition dataset
        """
        return self._dataset

    @dataset.setter
    def dataset(self, value):
        """
        Setter of attribute `self.dataset`

        Parameters
        ----------
        value: xarray.Dataset
            new Dataset
        """
        self._dataset = value

    @property
    def acquisition_type(self):
        """
        Gives the acquisition type (swath, truncated_swath,daily_regular_grid, model_regular_grid)

        Returns
        -------
        str
            acquisition type

        """
        return 'daily_regular_grid'

    @property
    def start_date(self):
        """
        Start acquisition time

        Returns
        -------
        numpy.datetime64
            Start time
        """
        return min(np.unique(self.dataset[self.time_name]))

    @property
    def stop_date(self):
        """
        Stop acquisition time

        Returns
        -------
        numpy.datetime64
            Stop time
        """
        return max(np.unique(self.dataset[self.time_name]))

    @property
    def orbit_segment_name(self):
        """
        Gives the name of the variable for orbit segmentation in dataset (Ascending / Descending). If value is None,
        so the orbit hasn't orbited segmentation

        Returns
        -------
        str | None
            Orbit segmentation variable name in the dataset. None if there isn't one.
        """
        return 'node'

    @property
    def has_orbited_segmentation(self):
        """
        True if there is orbit segmentation in the dataset

        Returns
        -------
        bool
            Presence or not of an orbit segmentation
        """
        if self.orbit_segment_name is not None:
            return True
        else:
            return False

    @property
    def minute_name(self):
        """
        Get name of the minute variable in the dataset

        Returns
        -------
        str
            Minute variable name
        """
        return 'minute'

    @property
    def day_date(self):
        """
        Get day date from the product name as a datetime

        Returns
        -------
        datetime.datetime
        Day date of the product
        """
        split_name = self.product_name.split('_')
        str_date = ''.join(split_name[4: 7])
        return datetime.strptime(str_date, '%Y%m%d')

    @property
    def mission_name(self):
        """
        Name of the mission (or model)

        Returns
        -------
        str
            Mission name (ex: SMOS, S1, RS2, RCM, SMAP, HY2, ERA5)
        """
        return "SMAP"

    def rename_vars_in_coloc(self, dataset=None):
        """
        Rename variables from a dataset to homogenize the co-location product. If no dataset is explicit, so it is this
        of `self.dataset` which is used.

        Parameters
        ----------
        dataset: xarray.Dataset | None
            Dataset on which common vars must be renamed

        Returns
        -------
        xarray.Dataset
            Dataset with homogene variable names
        """
        if dataset is None:
            dataset = self.dataset
        # map the variable names in the dataset with the keys in common vars
        mapper = {
            self.wind_name: 'wind_speed',
        }
        for var in dataset.variables:
            if var in mapper.keys():
                key_in_common_vars = mapper[var]
                dataset = dataset.rename_vars({var: common_var_names[key_in_common_vars]})
        return dataset

    @property
    def unecessary_vars_in_coloc_product(self):
        """
        Get unecessary variables in co-location product

        Returns
        -------
        list[str]
            Unecessary variables in co-location product
        """
        return [self.orbit_segment_name, self.time_name]

    @property
    def necessary_attrs_in_coloc_product(self):
        """
        Get necessary dataset attributes in co-location product

        Returns
        -------
        list[str]
            Necessary dataset attributes in co-location product
        """
        return ['Conventions', 'title', 'institution', 'grid_mapping', 'version', 'sourceReference']

    def rename_attrs_in_coloc_product(self, attr):
        """
        Get the new name of an attribute in co-location products from an original attribute

        Parameters
        ----------
        attr: str
            Attribute from the satellite dataset that needs to be renames for the co-location product.

        Returns
        -------
        str
            New attribute's name from the satellite dataset.
        """
        mapper = {
            'version': 'sourceProductVersion',
        }
        if attr in mapper.keys():
            return mapper[attr]
        else:
            return attr

    @property
    def wind_name(self):
        """
        Name of an important wind variable in the dataset

        Returns
        -------
        str
            Wind variable name

        """
        return 'wind'

    def add_source_reference_attribute(self, ds=None, attr_name='reference'):
        """
        Add the source reference attribute in a SMAP dataset. The name given to this attribute is chosen with the
        argument  `attr_name`

        Parameters
        ----------
        ds: xarray.Dataset | None
            Dataset that contains the SMAP information, in which the source reference attribute must be added.
        attr_name: str
            Name chosen for the reference attribute.

        Returns
        -------
        xarray.Dataset
            Dataset that contains the source reference attribute
        """
        content = {
            'authors': None,
            'year': None,
            'product': None,
            'institution': None,
            'weblink': None,
        }
        prefix = 'dataset_citation_'
        if ds is None:
            ds = self.dataset

        def is_empty():
            for k, val in content.items():
                if val is not None:
                    return False
                else:
                    content[k] = ''
            return True

        for key in content.keys():
            try:
                content[key] = ds.attrs[f"{prefix}{key}"]
            except KeyError:
                pass
        if is_empty():
            ds.attrs[attr_name] = ''
        else:
            version = 'Version 01.0. [NRT or FINAL].'
            ds.attrs[attr_name] = f"{content['authors']}, {content['year']}: {content['product']}, {version}, " + \
                                  f"{content['institution']}. {content['weblink']}"
        return ds

    @longitude_name.setter
    def longitude_name(self, value):
        self._longitude_name = value

    @latitude_name.setter
    def latitude_name(self, value):
        self._latitude_name = value
